"""
Functions to generate unique light cone solutions.

Author: Britton Smith <brittons@origins.colorado.edu>
Affiliation: CASA/University of CO, Boulder
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2008-2009 Britton Smith.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from light_cone import *
from Common_nVolume import *
from yt.logger import lagosLogger as mylog
import numpy as na
import random as rand
import copy
import sys

def project_unique_light_cones(lightcone, seed_file, field, **kwargs):
    "Make light cone projections using list of random seeds in a file."

    seedList = _read_seed_file(seed_file)

    prefix = lightcone.output_prefix
    lightcone.calculate_light_cone_solution(seed=0)
    lastSeed = None

    for seed in seedList:
        if (seed['master'] != lastSeed):
            lightcone.rerandomize_light_cone_solution(seed['master'], recycle=False)
            lastSeed = seed['master']
        if (seed['recycle'] is not None):
            lightcone.rerandomize_light_cone_solution(seed['recycle'], recycle=True)

        lightcone.output_prefix = "%s_%s_%s" % (prefix, seed['master'], seed['recycle'])
        lightcone.project_light_cone(field, **kwargs)

def find_unique_solutions(lightcone1, solutions=100, seed=None, max_overlap=0.25, failures=10, 
                          recycle=True, filename='unique.dat'):
    "Find a set of random seeds that will give light cones will minimal volume overlap."

    lightcone2 = copy.deepcopy(lightcone1)
    lightcone1.calculate_light_cone_solution(seed=0)
    lightcone2.calculate_light_cone_solution(seed=0)

    uniqueSeeds = []
    if recycle:
        master = None
    newRecycleSeed = None
    fails = 0
    recycleFails = 0

    maxCommon = 0.0

    # Need to continuall save and reset the state of the random number generator
    # since it is being reset by the light cone generator.
    if seed is None:
        state = None
    else:
        rand.seed(seed)
        state = rand.getstate()

    failDigits = str(int(na.log10(failures))+1)

    while (len(uniqueSeeds) < solutions):
        # Create new random seed.
        if (recycle and master is not None):
            newSeed = master
            if state is not None: rand.setstate(state)
            newRecycleSeed = rand.randint(1, 1e9)
            state = rand.getstate()
        else:
            if state is not None: rand.setstate(state)
            newSeed = rand.randint(1, 1e9)
            state = rand.getstate()
            if recycle:
                master = newSeed
                recycleFails = 0
            newRecycleSeed = None

        sys.stderr.write(("Unique solutions: %d, consecutive failures: %"+failDigits+"d, %"+failDigits+"d.\r") % \
                             (len(uniqueSeeds), fails, recycleFails))

        lightcone1.rerandomize_light_cone_solution(newSeed, recycle=False)
        if newRecycleSeed is not None:
            lightcone1.rerandomize_light_cone_solution(newRecycleSeed, recycle=True)

        # Compare with all other seeds.
        testPass = True
        for uniqueSeed in uniqueSeeds:
            lightcone2.rerandomize_light_cone_solution(uniqueSeed['master'], recycle=False)
            if uniqueSeed['recycle'] is not None:
                lightcone2.rerandomize_light_cone_solution(uniqueSeed['recycle'], recycle=True)

            common = _compare_solutions(lightcone1.light_cone_solution, lightcone2.light_cone_solution)

            if (common > max_overlap):
                testPass = False
                break
            else:
                maxCommon = max(maxCommon, common)

        if testPass:
            uniqueSeeds.append({'master':newSeed, 'recycle':newRecycleSeed})
            fails = 0
            recycleFails = 0

        else:
            if recycle:
                recycleFails += 1
            else:
                fails += 1

            if (recycleFails >= failures):
                sys.stderr.write(("Unique solutions: %d, consecutive failures: %"+failDigits+"d, %"+failDigits+"d.\n") % \
                                     (len(uniqueSeeds), fails, recycleFails))
                fails += 1
                mylog.info("Max recycled failures reached with master seed %d." % newSeed)
                master = None
            if (fails >= failures):
                sys.stderr.write(("Unique solutions: %d, consecutive failures: %"+failDigits+"d, %"+failDigits+"d.\n") % \
                                     (len(uniqueSeeds), fails, recycleFails))
                mylog.error("Max consecutive failures reached.")
                break

    mylog.info("Created %d unique solutions." % len(uniqueSeeds))
    mylog.info("Maximum common volume is %.2e." % maxCommon)
    _write_seed_file(uniqueSeeds, filename)
    return uniqueSeeds

def _compare_solutions(solution1, solution2):
    "Calculate common volume between two light cone solutions."

    if (len(solution1) != len(solution2)):
        mylog.error("Cannot compare light cone solutions with unequal numbers of slices.")
        return -1

    commonVolume = 0.0
    totalVolume = 0.0

    # Check that solution volumes are the same.
    if((solution1[0]['DepthBoxFraction'] * solution1[0]['WidthBoxFraction']**2) !=
       (solution2[0]['DepthBoxFraction'] * solution2[0]['WidthBoxFraction']**2)):
        mylog.error("Light cone solutions do not have equal volumes, will use the smaller one.")

    for q in range(len(solution1)):
        cube1 = na.zeros(shape=(len(solution1[q]['ProjectionCenter']), 2))
        volume1 = 1.0
        for w in range(len(cube1)):
            if (w == solution1[q]['ProjectionAxis']):
                width = solution1[q]['DepthBoxFraction']
            else:
                width = solution1[q]['WidthBoxFraction']
            volume1 *= width
            cube1[w] = [solution1[q]['ProjectionCenter'][w] - 0.5 * width,
                        solution1[q]['ProjectionCenter'][w] + 0.5 * width]

        cube2 = na.zeros(shape=(len(solution2[q]['ProjectionCenter']), 2))
        volume2 = 1.0
        for w in range(len(cube2)):
            if (w == solution2[q]['ProjectionAxis']):
                width = solution2[q]['DepthBoxFraction']
            else:
                width = solution2[q]['WidthBoxFraction']
            volume2 *= width
            cube2[w] = [solution2[q]['ProjectionCenter'][w] - 0.5 * width,
                        solution2[q]['ProjectionCenter'][w] + 0.5 * width]

        totalVolume += min(volume1, volume2)
        commonVolume += commonNVolume(cube1, cube2, periodic=na.array([[0, 1], [0, 1], [0, 1]]))

    return (commonVolume/totalVolume)

def _read_seed_file(filename):
    "Read list of random seeds from a file."

    mylog.info("Reading random seed list from %s." % filename)

    seedList = []

    lines = file(filename)
    for line in lines:
        if line[0] != '#':
            line = line.strip()
            onLine = line.split(',')
            if (len(onLine) == 1):
                seedList.append({'master':onLine[0], 'recycle':None})
            else:
                seedList.append({'master':onLine[0], 'recycle':onLine[1]})

    return seedList

@rootonly
def _write_seed_file(seedList, filename):
    "Write list of random seeds to a file."

    mylog.info("Writing random seed list to %s." % filename)

    f = open(filename, 'w')
    for seed in seedList:
        if seed['recycle'] is None:
            f.write("%s\n" % seed['master'])
        else:
            f.write("%s, %s\n" % (seed['master'], seed['recycle']))
    f.close()



"""
Quantities that can be derived from Enzo data that may also required additional
arguments.  (Standard arguments -- such as the center of a distribution of
points -- are excluded here, and left to the EnzoDerivedFields.)

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2007-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.lagos import *
from yt.funcs import get_pbar, wraps
import math

__CUDA_BLOCK_SIZE = 256

quantity_info = {}

class GridChildMaskWrapper:
    def __init__(self, grid, data_source):
        self.grid = grid
        self.data_source = data_source
    def __getattr__(self, attr):
        return getattr(self.grid, attr)
    def __getitem__(self, item):
        return self.data_source._get_data_from_grid(self.grid, item)

class DerivedQuantity(ParallelAnalysisInterface):
    def __init__(self, collection, name, function,
                 combine_function, units = "",
                 n_ret = 0, force_unlazy=False):
        # We wrap the function with our object
        self.__doc__ = function.__doc__
        self.__name__ = name
        self.collection = collection
        self._data_source = collection.data_source
        self.func = function
        self.c_func = combine_function
        self.n_ret = n_ret
        self.force_unlazy = force_unlazy

    def __call__(self, *args, **kwargs):
        lazy_reader = kwargs.pop('lazy_reader', True)
        preload = kwargs.pop('preload', False)
        if preload:
            if not lazy_reader: mylog.debug("Turning on lazy_reader because of preload")
            lazy_reader = True
            e = FieldDetector()
            e.NumberOfParticles = 1
            self.func(e, *args, **kwargs)
            mylog.debug("Preloading %s", e.requested)
            self._preload([g for g in self._get_grid_objs()], e.requested,
                          self._data_source.pf.h.io)
        if lazy_reader and not self.force_unlazy:
            return self._call_func_lazy(args, kwargs)
        else:
            return self._call_func_unlazy(args, kwargs)

    def _call_func_lazy(self, args, kwargs):
        self.retvals = [ [] for i in range(self.n_ret)]
        for gi,g in enumerate(self._get_grids()):
            rv = self.func(GridChildMaskWrapper(g, self._data_source), *args, **kwargs)
            for i in range(self.n_ret): self.retvals[i].append(rv[i])
            g.clear_data()
        self.retvals = [na.array(self.retvals[i]) for i in range(self.n_ret)]
        return self.c_func(self._data_source, *self.retvals)

    def _finalize_parallel(self):
        self.retvals = [na.array(self._mpi_catlist(my_list)) for my_list in self.retvals]
        
    def _call_func_unlazy(self, args, kwargs):
        retval = self.func(self._data_source, *args, **kwargs)
        return self.c_func(self._data_source, *retval)

def add_quantity(name, **kwargs):
    if 'function' not in kwargs or 'combine_function' not in kwargs:
        mylog.error("Not adding field %s because both function and combine_function must be provided" % name)
        return
    f = kwargs.pop('function')
    c = kwargs.pop('combine_function')
    quantity_info[name] = (name, f, c, kwargs)

class DerivedQuantityCollection(object):
    functions = quantity_info
    def __init__(self, data_source):
        self.data_source = data_source

    def __getitem__(self, key):
        if key not in self.functions:
            raise KeyError(key)
        args = self.functions[key][:3]
        kwargs = self.functions[key][3]
        # Instantiate here, so we can pass it the data object
        # Note that this means we instantiate every time we run help, etc
        # I have made my peace with this.
        return DerivedQuantity(self, *args, **kwargs)

    def keys(self):
        return self.functions.keys()

def _TotalMass(data):
    """
    This function takes no arguments and returns the sum of cell masses and
    particle masses in the object.
    """
    baryon_mass = data["CellMassMsun"].sum()
    particle_mass = data["ParticleMassMsun"].sum()
    return baryon_mass, particle_mass
def _combTotalMass(data, baryon_mass, particle_mass):
    return baryon_mass.sum() + particle_mass.sum()
add_quantity("TotalMass", function=_TotalMass,
             combine_function=_combTotalMass, n_ret = 2)

def _CenterOfMass(data):
    """
    This function takes no arguments and returns the location of the center
    of mass of the *non-particle* data in the object.
    """
    x = (data["x"] * data["CellMassMsun"]).sum()
    y = (data["y"] * data["CellMassMsun"]).sum()
    z = (data["z"] * data["CellMassMsun"]).sum()
    den = data["CellMassMsun"].sum()
    return x,y,z, den
def _combCenterOfMass(data, x,y,z, den):
    return na.array([x.sum(), y.sum(), z.sum()])/den.sum()
add_quantity("CenterOfMass", function=_CenterOfMass,
             combine_function=_combCenterOfMass, n_ret = 4)

def _WeightedAverageQuantity(data, field, weight):
    """
    This function returns an averaged quantity.

    :param field: The field to average
    :param weight: The field to weight by
    """
    num = (data[field] * data[weight]).sum()
    den = data[weight].sum()
    return num, den
def _combWeightedAverageQuantity(data, field, weight):
    return field.sum()/weight.sum()
add_quantity("WeightedAverageQuantity", function=_WeightedAverageQuantity,
             combine_function=_combWeightedAverageQuantity, n_ret = 2)

def _BulkVelocity(data):
    """
    This function returns the mass-weighted average velocity in the object.
    """
    xv = (data["x-velocity"] * data["CellMassMsun"]).sum()
    yv = (data["y-velocity"] * data["CellMassMsun"]).sum()
    zv = (data["z-velocity"] * data["CellMassMsun"]).sum()
    w = data["CellMassMsun"].sum()
    return xv, yv, zv, w
def _combBulkVelocity(data, xv, yv, zv, w):
    w = w.sum()
    xv = xv.sum()/w
    yv = yv.sum()/w
    zv = zv.sum()/w
    return na.array([xv, yv, zv])
add_quantity("BulkVelocity", function=_BulkVelocity,
             combine_function=_combBulkVelocity, n_ret=4)

def _AngularMomentumVector(data):
    """
    This function returns the mass-weighted average angular momentum vector.
    """
    am = data["SpecificAngularMomentum"]*data["CellMassMsun"]
    j_mag = am.sum(axis=1)
    return [j_mag]
def _combAngularMomentumVector(data, j_mag):
    if len(j_mag.shape) < 2: j_mag = na.expand_dims(j_mag, 0)
    L_vec = j_mag.sum(axis=0)
    L_vec_norm = L_vec / na.sqrt((L_vec**2.0).sum())
    return L_vec_norm
add_quantity("AngularMomentumVector", function=_AngularMomentumVector,
             combine_function=_combAngularMomentumVector, n_ret=1)

def _BaryonSpinParameter(data):
    """
    This function returns the spin parameter for the baryons, but it uses
    the particles in calculating enclosed mass.
    """
    m_enc = data["CellMassMsun"].sum() + data["ParticleMassMsun"].sum()
    am = data["SpecificAngularMomentum"]*data["CellMassMsun"]
    j_mag = am.sum(axis=1)
    e_term_pre = na.sum(data["CellMassMsun"]*data["VelocityMagnitude"]**2.0)
    weight=data["CellMassMsun"].sum()
    return j_mag, m_enc, e_term_pre, weight
def _combBaryonSpinParameter(data, j_mag, m_enc, e_term_pre, weight):
    # Because it's a vector field, we have to ensure we have enough dimensions
    if len(j_mag.shape) < 2: j_mag = na.expand_dims(j_mag, 0)
    W = weight.sum()
    M = m_enc.sum()
    J = na.sqrt(((j_mag.sum(axis=0))**2.0).sum())/W
    E = na.sqrt(e_term_pre.sum()/W)
    G = 6.67e-8 # cm^3 g^-1 s^-2
    spin = J * E / (M*1.989e33*G)
    return spin
add_quantity("BaryonSpinParameter", function=_BaryonSpinParameter,
             combine_function=_combBaryonSpinParameter, n_ret=4)

def _ParticleSpinParameter(data):
    """
    This function returns the spin parameter for the baryons, but it uses
    the particles in calculating enclosed mass.
    """
    m_enc = data["CellMassMsun"].sum() + data["ParticleMassMsun"].sum()
    am = data["ParticleSpecificAngularMomentum"]*data["ParticleMassMsun"]
    if am.size == 0: return (na.zeros((3,), dtype='float64'), m_enc, 0, 0)
    j_mag = am.sum(axis=1)
    e_term_pre = na.sum(data["ParticleMassMsun"]
                       *data["ParticleVelocityMagnitude"]**2.0)
    weight=data["ParticleMassMsun"].sum()
    return j_mag, m_enc, e_term_pre, weight
add_quantity("ParticleSpinParameter", function=_ParticleSpinParameter,
             combine_function=_combBaryonSpinParameter, n_ret=4)
    
def _IsBound(data, truncate = True, include_thermal_energy = False):
    """
    This returns whether or not the object is gravitationally bound
    
    :param truncate: Should the calculation stop once the ratio of
                     gravitational:kinetic is 1.0?
    :param include_thermal_energy: Should we add the energy from ThermalEnergy
                                   on to the kinetic energy to calculate 
                                   binding energy?
    """
    # Kinetic energy
    bv_x,bv_y,bv_z = data.quantities["BulkVelocity"]()
    # One-cell objects are NOT BOUND.
    if data["CellMass"].size == 1: return [0.0]
    kinetic = 0.5 * (data["CellMass"] * (
                       (data["x-velocity"] - bv_x)**2
                     + (data["y-velocity"] - bv_y)**2
                     + (data["z-velocity"] - bv_z)**2 )).sum()
    # Add thermal energy to kinetic energy
    if (include_thermal_energy):
        thermal = (data["ThermalEnergy"] * data["CellMass"]).sum()
        kinetic += thermal
    # Gravitational potential energy
    # We only divide once here because we have velocity in cgs, but radius is
    # in code.
    G = 6.67e-8 / data.convert("cm") # cm^3 g^-1 s^-2
    import time
    t1 = time.time()
    try:
        pot = G*_cudaIsBound(data, truncate, kinetic/G)
    except (ImportError, AssertionError):
        pot = G*PointCombine.FindBindingEnergy(data["CellMass"],
                                               data['x'],data['y'],data['z'],
                                               truncate, kinetic/G)
    mylog.info("Boundedness check took %0.3e seconds", time.time()-t1)
    return [(pot / kinetic)]
def _combIsBound(data, bound):
    return bound
add_quantity("IsBound",function=_IsBound,combine_function=_combIsBound,n_ret=1,
             force_unlazy=True)

def _cudaIsBound(data, truncate, ratio):
    bsize = __CUDA_BLOCK_SIZE
    import pycuda.driver as cuda
    import pycuda.autoinit
    import pycuda.gpuarray as gpuarray
    my_stream = cuda.Stream()
    cuda.init()
    assert cuda.Device.count() >= 1

    mass_scale_factor = 1.0/(data['CellMass'].max())
    m = (data['CellMass'] * mass_scale_factor).astype('float32')
    assert(m.size > bsize)

    gsize=int(math.ceil(float(m.size)/bsize))
    assert(gsize > 16)

    # Now the tedious process of rescaling our values...
    length_scale_factor = data['dx'].max()/data['dx'].min()
    x = ((data['x'] - data['x'].min()) * length_scale_factor).astype('float32')
    y = ((data['y'] - data['y'].min()) * length_scale_factor).astype('float32')
    z = ((data['z'] - data['z'].min()) * length_scale_factor).astype('float32')
    p = na.zeros(z.shape, dtype='float32')
    
    x_gpu = cuda.mem_alloc(x.size * x.dtype.itemsize)
    y_gpu = cuda.mem_alloc(y.size * y.dtype.itemsize)
    z_gpu = cuda.mem_alloc(z.size * z.dtype.itemsize)
    m_gpu = cuda.mem_alloc(m.size * m.dtype.itemsize)
    p_gpu = cuda.mem_alloc(p.size * p.dtype.itemsize)
    for ag, a in [(x_gpu, x), (y_gpu, y), (z_gpu, z), (m_gpu, m), (p_gpu, p)]:
        cuda.memcpy_htod(ag, a)
    source = """

      extern __shared__ float array[];

      __global__ void isbound(float *x, float *y, float *z, float *m,
                              float *p, int *nelem)
      {

        /* My index in the array */
        int idx1 = blockIdx.x * blockDim.x + threadIdx.x;
        /* Note we are setting a start index */
        int idx2 = blockIdx.y * blockDim.x;
        int offset = threadIdx.x;

        /* Here we're just setting up convenience pointers to our
           shared array */

        float* x_data1 = (float*) array;
        float* y_data1 = (float*) &x_data1[blockDim.x];
        float* z_data1 = (float*) &y_data1[blockDim.x];
        float* m_data1 = (float*) &z_data1[blockDim.x];

        float* x_data2 = (float*) &m_data1[blockDim.x];
        float* y_data2 = (float*) &x_data2[blockDim.x];
        float* z_data2 = (float*) &y_data2[blockDim.x];
        float* m_data2 = (float*) &z_data2[blockDim.x];

        x_data1[offset] = x[idx1];
        y_data1[offset] = y[idx1];
        z_data1[offset] = z[idx1];
        m_data1[offset] = m[idx1];

        x_data2[offset] = x[idx2 + offset];
        y_data2[offset] = y[idx2 + offset];
        z_data2[offset] = z[idx2 + offset];
        m_data2[offset] = m[idx2 + offset];

        __syncthreads();

        float tx, ty, tz;

        float my_p = 0.0;

        if(idx1 < %(p)s) {
            for (int i = 0; i < blockDim.x; i++){
                if(i + idx2 < idx1 + 1) continue;
                tx = (x_data1[offset]-x_data2[i]);
                ty = (y_data1[offset]-y_data2[i]);
                tz = (z_data1[offset]-z_data2[i]);
                my_p += m_data1[offset]*m_data2[i] /
                    sqrt(tx*tx+ty*ty+tz*tz);
            }
        }
        p[idx1] += my_p;
        __syncthreads();
      }
    """
    mod = cuda.SourceModule(source % dict(p=m.size))
    func = mod.get_function('isbound')
    mylog.info("Running CUDA functions.  May take a while.  (%0.5e, %s)",
               x.size, gsize)
    import pycuda.tools as ct
    t1 = time.time()
    ret = func(x_gpu, y_gpu, z_gpu, m_gpu, p_gpu,
               shared=8*bsize*m.dtype.itemsize,
         block=(bsize,1,1), grid=(gsize, gsize), time_kernel=True)
    cuda.memcpy_dtoh(p, p_gpu)
    p1 = p.sum()
    if na.any(na.isnan(p)): raise ValueError
    return p1 * (length_scale_factor / (mass_scale_factor**2.0))

def _Extrema(data, fields, non_zero = False, filter=None):
    """
    This function returns the extrema of a set of fields
    
    :param fields: A field name, or a list of field names
    :param filter: a string to be evaled to serve as a data filter.
    """
    # There is a heck of a lot of logic in this.  I really wish it were more
    # elegant.
    fields = ensure_list(fields)
    if filter is not None: this_filter = eval(filter)
    mins, maxs = [], []
    for field in fields:
        if data[field].size < 1:
            mins.append(1e90)
            maxs.append(-1e90)
            continue
        if filter is None:
            if non_zero:
                nz_filter = data[field]>0.0
                if not nz_filter.any():
                    mins.append(1e90)
                    maxs.append(-1e90)
                    continue
            else:
                nz_filter = None
            mins.append(data[field][nz_filter].min())
            maxs.append(data[field][nz_filter].max())
        else:
            if this_filter.any():
                if non_zero:
                    nz_filter = ((this_filter) &
                                 (data[field][this_filter] > 0.0))
                else: nz_filter = this_filter
                mins.append(data[field][nz_filter].min())
                maxs.append(data[field][nz_filter].max())
            else:
                mins.append(1e90)
                maxs.append(-1e90)
    return len(fields), mins, maxs
def _combExtrema(data, n_fields, mins, maxs):
    mins, maxs = na.atleast_2d(mins, maxs)
    n_fields = mins.shape[1]
    return [(na.min(mins[:,i]), na.max(maxs[:,i])) for i in range(n_fields)]
add_quantity("Extrema", function=_Extrema, combine_function=_combExtrema,
             n_ret=3)

def _Action(data, action, combine_action, filter=None):
    """
    This function evals the string given by the action arg and uses 
    the function thrown with the combine_action to combine the values.  
    A filter can be thrown to be evaled to short-circuit the calculation 
    if some criterion is not met.
    :param action: a string containing the desired action to be evaled.
    :param combine_action: the function used to combine the answers when done lazily.
    :param filter: a string to be evaled to serve as a data filter.
    """
    if filter is not None:
        if not eval(filter).any(): return 0, False, combine_action
    value = eval(action)
    return value, True, combine_action
def _combAction(data, value, valid, combine_action):
    return combine_action[0](value[valid])
add_quantity("Action", function=_Action, combine_function=_combAction, n_ret=3)

def _MaxLocation(data, field):
    """
    This function returns the location of the maximum of a set
    of fields.
    """
    ma, maxi, mx, my, mz, mg = -1e90, -1, -1, -1, -1, -1
    if data[field].size > 0:
        maxi = na.argmax(data[field])
        ma = data[field][maxi]
        mx, my, mz = [data[ax][maxi] for ax in 'xyz']
        mg = data["GridIndices"][maxi]
    return (ma, maxi, mx, my, mz, mg)
def _combMaxLocation(data, *args):
    args = [na.atleast_1d(arg) for arg in args]
    i = na.argmax(args[0]) # ma is arg[0]
    return [arg[i] for arg in args]
add_quantity("MaxLocation", function=_MaxLocation,
             combine_function=_combMaxLocation, n_ret = 6)

def _TotalQuantity(data, fields):
    """
    This function sums up a given field over the entire region

    :param fields: The fields to sum up
    """
    fields = ensure_list(fields)
    totals = []
    for field in fields:
        if data[field].size < 1:
            totals.append(0)
            continue
        totals.append(data[field].sum())
    return len(fields), totals
def _combTotalQuantity(data, n_fields, totals):
    totals = na.atleast_2d(totals)
    n_fields = totals.shape[1]
    return [na.sum(totals[:,i]) for i in range(n_fields)]
add_quantity("TotalQuantity", function=_TotalQuantity,
                combine_function=_combTotalQuantity, n_ret=2)



#!python2.5

import optparse
import math
import os, os.path
import sys

my_opts = optparse.OptionParser()
my_opts.add_option("-w", "--width",
                   action="store", type="float",
                   dest="width", default=1.0,
                   help="Width in specified units")
my_opts.add_option("-u", "--unit",
                   action="store", type="string",
                   dest="unit", default='1',
                   help="Desired units")
my_opts.add_option("-k", "--slab-width",
                   action="store", type="float",
                   dest="slab_width", default=1.0,
                   help="Slab width in specified units")
my_opts.add_option("-g", "--slab-unit",
                   action="store", type="string",
                   dest="slab_unit", default='1',
                   help="Desired units for the slab")
my_opts.add_option("-b", "--basename",
                   action="store", type="string",
                   dest="basename", default="galaxy",
                   help="Basename of parameter files")
my_opts.add_option("-c", "--center",
                   action="store", type="float",
                   dest="center", default=None,
                   nargs=3,
                   help="Center (-1,-1,-1 for max)")
my_opts.add_option("-z", "--zlim",
                   action="store", type="float",
                   dest="zlim", default=None,
                   nargs=2,
                   help="Color limits (min, max)")
my_opts.add_option("-a", "--axis",
                   action="store", type="int",
                   dest="axis", default=4,
                   help="Axis (4 for all three)")
my_opts.add_option("-l", "--log",
                   action="store_true",
                   dest="takelog", default=False,
                   help="Take the log of the field?")
my_opts.add_option("-f", "--field",
                   action="store", type="string",
                   dest="field", default=None,
                   help="Field to color by")
my_opts.add_option("-s", "--skip",
                   action="store", type="int",
                   dest="skip", default=1,
                   help="Skip factor for outputs")
my_opts.add_option("-m", "--colormap",
                   action="store", type="string",
                   dest="cmap", default="jet",
                   help="Colormap name")
my_opts.add_option("-o", "--output",
                   action="store", type="string",
                   dest="output", default="particle_frames/",
                   help="Folder in which to place output frames")
my_opts.add_option("-t", "--particle-type",
                   action="store", type="int",
                   dest="ptype", default=2,
                   help="Particle type to select")
my_opts.add_option("-e", "--age-cut",
                   action="store", type="float",
                   dest="age_filter", default=None,
                   nargs=2,
                   help="Bounds for the field to select")


opts, args = my_opts.parse_args()

import yt.lagos as lagos
import yt.raven as raven
import yt.fido as fido
import numpy as na
mylog = fido.mylog



if not os.path.isdir(opts.output):
    os.mkdir(opts.output)

try:
    first = int(args[0])
    last = int(args[1])
except:
    mylog.error("Hey, sorry, but you need to specify the first and last outputs you want to look at.")
    sys.exit()

print first, last, range(first, last+1, opts.skip)

import pylab
from matplotlib.colors import LogNorm, Normalize

if opts.takelog:
    myNorm = LogNorm
else:
    myNorm = Normalize

for n in range(first,last+1,opts.skip): # This is the array of galaxy outputs we want
    # Now we figure out where this file is
    bn_try = opts.basename + "%04i" % n
    if os.path.isfile(bn_try):
        fn = bn_try
    else:
        fn = os.path.join(bn_try + ".dir", bn_try)
    mylog.info("Now attempting to zoom in on %s", fn)
    try:
        a = lagos.EnzoStaticOutput(fn)
        min_dx = a.h.get_smallest_dx()
    except:
        mylog.warning("Something messed up!  Are you sure you gave good info?")
        continue

    if opts.center == None:
        mylog.info("No center fed in; seeking.")
        v, center = a.h.find_max("Density")
    else:
        mylog.info("Center fed in; not seeking.")
        center = opts.center
    mylog.info("Setting center to %0.3f %0.3f %0.3f", 
                center[0], center[1], center[2])

    # Now we deal with all the plotting.

    if opts.axis == 4:
        axes = range(3)
    else:
        axes = [opts.axis]

    gI = na.where(a.h.gridNumberOfParticles.ravel() > 0)

    npart = na.sum(a.h.gridNumberOfParticles)
    pos = na.zeros((npart,4)) # three plus fielD
    ni = 0
    mylog.info("Getting particles from %s grids", len(gI[0]))
    for grid in a.h.grids[gI]:
        pi = na.where(grid["particle_type"] == opts.ptype)
        x = grid["particle_position_x"][pi]
        y = grid["particle_position_y"][pi]
        z = grid["particle_position_z"][pi]
        pos[ni:ni+x.shape[0],0] = x
        pos[ni:ni+x.shape[0],1] = y
        pos[ni:ni+x.shape[0],2] = z
        if opts.field:
            if opts.field == "age":
                pos[ni:ni+x.shape[0],3] = \
                  (a["InitialTime"] - grid["creation_time"][pi]) \
                  * a["years"]
            else:
                pos[ni:ni+x.shape[0],3] = grid[opts.field][pi]
        ni += x.shape[0]

    for ax in axes:
        mylog.info("Adding plot for axis %i", ax)
        pylab.clf()
        xa = lagos.x_dict[ax]
        ya = lagos.y_dict[ax]
        
        z1 = center[ax] - opts.slab_width/a[opts.slab_unit] * 0.5
        z2 = center[ax] + opts.slab_width/a[opts.slab_unit] * 0.5

        newpos = pos[na.where((pos[:,ax] > z1)
                            & (pos[:,ax] < z2))]
        if newpos.shape[0] == 0:
            mylog.info("Hey, no particles match the criteria.  Moving on.")
            continue
        if opts.age_filter and opts.field == "age":
            newpos = newpos[na.where((newpos[:,3] > opts.age_filter[0])
                                   & (newpos[:,3] < opts.age_filter[1]))]
        mylog.info("Constraining to %s - %s, which gives us %s particles",
                    z1,z2,newpos.shape[0])
        if opts.field:
            #print "MM", \
                  #newpos[:,xa].min(), newpos[:,xa].max(), \
                  #newpos[:,3].min(), newpos[:,3].max()
            pylab.scatter(newpos[:,xa].ravel(),
                          newpos[:,ya].ravel(),
                          c=newpos[:,3].ravel(),
                          faceted=False,s=1.0,
                          alpha=1.0,
                          norm=myNorm())
            if opts.zlim:
                pylab.clim(opts.zlim[0], opts.zlim[1])
            pylab.colorbar()
        else:
            pylab.scatter(newpos[:,xa].ravel(),
                          newpos[:,ya].ravel(),
                          c='k',faceted=False,s=1.0,
                          alpha=1.0)
            
        x1 = center[xa] - \
             opts.width / (2.0*a[opts.unit])
        x2 = center[xa] + \
             opts.width / (2.0*a[opts.unit])
        y1 = center[ya] - \
             opts.width / (2.0*a[opts.unit])
        y2 = center[ya] + \
             opts.width / (2.0*a[opts.unit])
        pylab.xlim(x1, x2)
        pylab.ylim(y1, y2)
        pylab.savefig(os.path.join(opts.output,"particle_frame%06i_%s.png"
                      % (n,lagos.axis_names[ax])))



"""
This is a definition of a class for providing data to a simulation
verification method

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2007-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.mods import *

class TestDataSource(object):
    def provide(self, pf):
        pass

    def __call__(self, pf):
        return self.provide(pf)

class MaximumDensitySphere(TestDataSource):
    def __init__(self, radius, unit):
        self.radius = radius
        self.unit = unit

    def provide(self, pf):
        v, c = pf.h.find_max("Density")
        yield pf.h.sphere(c, self.radius/pf[self.unit])

class HaloSpheres(TestDataSource):
    def __init__(self, n_part):
        self.n_part = n_part

    def provide(self, pf):
       halo_list = HaloFinder(pf)
       for halo in halo_list:
            if halo.get_size() < self.n_part: continue
            yield halo.get_sphere()

class Halos(TestDataSource):
    def __init__(self):
        pass

    def provide(self, pf):
       halo_list = HaloFinder(pf)
       for halo in halo_list:
            yield halo

class EntireDataset(TestDataSource):
    def __init__(self):
        pass

    def provide(self, pf):
        yield pf.h.all_data()



#!/usr/bin/env python
import setuptools

def configuration(parent_package='',top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('extensions',parent_package,top_path)
    config.make_config_py() # installs __config__.py
    config.make_svn_version_py()
    config.add_subpackage("lightcone")
    config.add_subpackage("volume_rendering")
    config.add_subpackage("kdtree")
    config.add_subpackage("image_panner")
    return config



"""
The data-file handling functions

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2007-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from yt.lagos import *
import yt.amr_utils as au
import exceptions

_axis_ids = {0:2,1:1,2:0}

io_registry = {}

class BaseIOHandler(object):

    _data_style = None
    _particle_reader = False

    class __metaclass__(type):
        def __init__(cls, name, b, d):
            type.__init__(cls, name, b, d)
            if hasattr(cls, "_data_style"):
                io_registry[cls._data_style] = cls

    def __init__(self):
        self.queue = defaultdict(dict)

    # We need a function for reading a list of sets
    # and a function for *popping* from a queue all the appropriate sets

    def preload(self, grids, sets):
        pass

    def pop(self, grid, field):
        if grid.id in self.queue and field in self.queue[grid.id]:
            return self.modify(self.queue[grid.id].pop(field))
        else:
            # We only read the one set and do not store it if it isn't pre-loaded
            return self._read_data_set(grid, field)

    def peek(self, grid, field):
        return self.queue[grid.id].get(field, None)

    def push(self, grid, field, data):
        if grid.id in self.queue and field in self.queue[grid.id]:
            raise ValueError
        self.queue[grid][field] = data

    # Now we define our interface
    def _read_data_set(self, grid, field):
        pass

    def _read_data_slice(self, grid, field, axis, coord):
        pass

    def _read_field_names(self, grid):
        pass

    @property
    def _read_exception(self):
        return None

class IOHandlerEnzoHDF4(BaseIOHandler):

    _data_style = "enzo_hdf4"

    def modify(self, field):
        return field.swapaxes(0,2)

    def _read_field_names(self, grid):
        """
        Returns a list of fields associated with the filename
        Should *only* be called as EnzoGridInstance.getFields, never as getFields(object)
        """
        return SD.SD(grid.filename).datasets().keys()

    def _read_data_set(self, grid, field):
        """
        Returns after having obtained or generated a field.  Should throw an
        exception.  Should only be called as EnzoGridInstance.readData()

        @param field: field to read
        @type field: string
        """
        return SD.SD(grid.filename).select(field).get()

    def _read_data_slice(self, grid, field, axis, coord):
        """
        Reads a slice through the HDF4 data

        @param grid: Grid to slice
        @type grid: L{EnzoGrid<EnzoGrid>}
        @param field: field to get
        @type field: string
        @param sl: region to get
        @type sl: SliceType
        """
        sl = [slice(None), slice(None), slice(None)]
        sl[axis] = slice(coord, coord + 1)
        sl = tuple(reversed(sl))
        return SD.SD(grid.filename).select(field)[sl].swapaxes(0,2)

    @property
    def _read_exception(self):
        return SD.HDF4Error

class IOHandlerEnzoHDF4_2D(IOHandlerEnzoHDF4):

    _data_style = "enzo_hdf4_2d"

    def _read_data_set(self, grid, field):
        t = SD.SD(grid.filename).select(field).get()[:,:,None]
        return t.swapaxes(0,1)

    def _read_data_slice(self, grid, field, axis, coord):
        t = SD.SD(grid.filename).select(field).get()
        return t.transpose()

    def modify(self, field):
        return field

class IOHandlerEnzoHDF5(BaseIOHandler):

    _data_style = "enzo_hdf5"
    _particle_reader = True

    def _read_field_names(self, grid):
        """
        Returns a list of fields associated with the filename
        Should *only* be called as EnzoGridInstance.getFields, never as getFields(object)
        """
        return HDF5LightReader.ReadListOfDatasets(grid.filename, "/")

    def _read_data_set(self, grid, field):
        return HDF5LightReader.ReadData(grid.filename, "/%s" % field).swapaxes(0,2)

    def _read_data_slice(self, grid, field, axis, coord):
        """
        Reads a slice through the HDF5 data

        @param grid: Grid to slice
        @type grid: L{EnzoGrid<EnzoGrid>}
        @param field: field to get
        @type field: string
        @param axis: axis to slice along
        @param coord: coord to slice at
        """
        axis = {0:2,1:1,2:0}[axis]
        t = HDF5LightReader.ReadDataSlice(grid.filename, "/%s" %
                        (field), axis, coord).transpose()
        return t

    def modify(self, field):
        return field.swapaxes(0,2)

    @property
    def _read_exception(self):
        return (exceptions.KeyError, HDF5LightReader.ReadingError)

    def _read_particles(self, fields, rtype, args, grid_list, enclosed,
                        conv_factors):
        filenames = [g.filename for g in grid_list]
        ids = [g.id for g in grid_list]
        return HDF5LightReader.ReadParticles(
            rtype, fields, filenames, ids, conv_factors, args, 0)

class IOHandlerExtracted(BaseIOHandler):

    _data_style = 'extracted'

    def _read_data_set(self, grid, field):
        return (grid.base_grid[field] / grid.base_grid.convert(field))

    def _read_data_slice(self, grid, field, axis, coord):
        sl = [slice(None), slice(None), slice(None)]
        sl[axis] = slice(coord, coord + 1)
        return grid.base_grid[field][tuple(sl)] / grid.base_grid.convert(field)

class IOHandlerPackedHDF5(BaseIOHandler):

    _data_style = "enzo_packed_3d"
    _particle_reader = True

    def _read_particles(self, fields, rtype, args, grid_list, enclosed,
                        conv_factors):
        filenames = [g.filename for g in grid_list]
        ids = [g.id for g in grid_list]
        filenames, ids, conv_factors = zip(*sorted(zip(filenames, ids, conv_factors)))
        return HDF5LightReader.ReadParticles(
            rtype, fields, list(filenames), list(ids), conv_factors, args, 1)

    def modify(self, field):
        return field.swapaxes(0,2)

    def preload(self, grids, sets):
        # We need to deal with files first
        files_keys = defaultdict(lambda: [])
        pf_field_list = grids[0].pf.h.field_list
        sets = [dset for dset in list(sets) if dset in pf_field_list]
        for g in grids: files_keys[g.filename].append(g)
        exc = self._read_exception
        for file in files_keys:
            mylog.debug("Starting read %s (%s)", file, sets)
            nodes = [g.id for g in files_keys[file]]
            nodes.sort()
            # We want to pass on any error we might expect -- the preload
            # phase should be non-fatal in all cases, and instead dump back to
            # the grids.
            data = HDF5LightReader.ReadMultipleGrids(file, nodes, sets)
            mylog.debug("Read %s items from %s", len(data), os.path.basename(file))
            for gid in data: self.queue[gid].update(data[gid])
        mylog.debug("Finished read of %s", sets)

    def _read_data_set(self, grid, field):
        return HDF5LightReader.ReadData(grid.filename,
                "/Grid%08i/%s" % (grid.id, field)).swapaxes(0,2)

    def _read_data_slice(self, grid, field, axis, coord):
        axis = _axis_ids[axis]
        return HDF5LightReader.ReadDataSlice(grid.filename, "/Grid%08i/%s" %
                        (grid.id, field), axis, coord).transpose()

    def _read_field_names(self, grid):
        return HDF5LightReader.ReadListOfDatasets(
                    grid.filename, "/Grid%08i" % grid.id)

    @property
    def _read_exception(self):
        return (exceptions.KeyError, HDF5LightReader.ReadingError)

class IOHandlerInMemory(BaseIOHandler):

    _data_style = "enzo_inline"

    def __init__(self, ghost_zones=3):
        import enzo
        self.enzo = enzo
        self.grids_in_memory = enzo.grid_data
        self.old_grids_in_memory = enzo.old_grid_data
        self.my_slice = (slice(ghost_zones,-ghost_zones),
                      slice(ghost_zones,-ghost_zones),
                      slice(ghost_zones,-ghost_zones))
        BaseIOHandler.__init__(self)

    def _read_data_set(self, grid, field):
        if grid.id not in self.grids_in_memory: raise KeyError
        tr = self.grids_in_memory[grid.id][field]
        # If it's particles, we copy.
        if len(tr.shape) == 1: return tr.copy()
        # New in-place unit conversion breaks if we don't copy first
        return tr.swapaxes(0,2)[self.my_slice].copy()
        # We don't do this, because we currently do not interpolate
        coef1 = max((grid.Time - t1)/(grid.Time - t2), 0.0)
        coef2 = 1.0 - coef1
        t1 = enzo.yt_parameter_file["InitialTime"]
        t2 = enzo.hierarchy_information["GridOldTimes"][grid.id]
        return (coef1*self.grids_in_memory[grid.id][field] + \
                coef2*self.old_grids_in_memory[grid.id][field])\
                [self.my_slice]

    def modify(self, field):
        return field.swapaxes(0,2)

    def _read_field_names(self, grid):
        return self.grids_in_memory[grid.id].keys()

    def _read_data_slice(self, grid, field, axis, coord):
        sl = [slice(3,-3), slice(3,-3), slice(3,-3)]
        sl[axis] = slice(coord + 3, coord + 4)
        sl = tuple(reversed(sl))
        tr = self.grids_in_memory[grid.id][field][sl].swapaxes(0,2)
        # In-place unit conversion requires we return a copy
        return tr.copy()

    @property
    def _read_exception(self):
        return KeyError

class IOHandlerNative(BaseIOHandler):

    _data_style = "orion_native"

    def modify(self, field):
        return field.swapaxes(0,2)

    def _read_data_set(self,grid,field):
        """
        reads packed multiFABs output by BoxLib in "NATIVE" format.

        """
        filen = os.path.expanduser(grid.filename[field])
        off = grid._offset[field]
        inFile = open(filen,'rb')
        inFile.seek(off)
        header = inFile.readline()
        header.strip()

        if grid._paranoid:
            mylog.warn("Orion Native reader: Paranoid read mode.")
            headerRe = re.compile(orion_FAB_header_pattern)
            bytesPerReal,endian,start,stop,centerType,nComponents = headerRe.search(header).groups()

            # we will build up a dtype string, starting with endian
            # check endianness (this code is ugly. fix?)
            bytesPerReal = int(bytesPerReal)
            if bytesPerReal == int(endian[0]):
                dtype = '<'
            elif bytesPerReal == int(endian[-1]):
                dtype = '>'
            else:
                raise ValueError("FAB header is neither big nor little endian. Perhaps the file is corrupt?")

            dtype += ('f%i'% bytesPerReal) #always a floating point

            # determine size of FAB
            start = na.array(map(int,start.split(',')))
            stop = na.array(map(int,stop.split(',')))

            gridSize = stop - start + 1

            error_count = 0
            if (start != grid.start).any():
                print "Paranoia Error: Cell_H and %s do not agree on grid start." %grid.filename
                error_count += 1
            if (stop != grid.stop).any():
                print "Paranoia Error: Cell_H and %s do not agree on grid stop." %grid.filename
                error_count += 1
            if (gridSize != grid.ActiveDimensions).any():
                print "Paranoia Error: Cell_H and %s do not agree on grid dimensions." %grid.filename
                error_count += 1
            if bytesPerReal != grid.hierarchy._bytesPerReal:
                print "Paranoia Error: Cell_H and %s do not agree on bytes per real number." %grid.filename
                error_count += 1
            if (bytesPerReal == grid.hierarchy._bytesPerReal and dtype != grid.hierarchy._dtype):
                print "Paranoia Error: Cell_H and %s do not agree on endianness." %grid.filename
                error_count += 1

            if error_count > 0:
                raise RunTimeError("Paranoia unveiled %i differences between Cell_H and %s." % (error_count, grid.filename))

        else:
            start = grid.start_index
            stop = grid.stop_index
            dtype = grid.hierarchy._dtype
            bytesPerReal = grid.hierarchy._bytesPerReal

        nElements = grid.ActiveDimensions.prod()

        # one field has nElements*bytesPerReal bytes and is located
        # nElements*bytesPerReal*field_index from the offset location
        if yt2orionFieldsDict.has_key(field):
            fieldname = yt2orionFieldsDict[field]
        else:
            fieldname = field
        field_index = grid.field_indexes[fieldname]
        inFile.seek(int(nElements*bytesPerReal*field_index),1)
        field = na.fromfile(inFile,count=nElements,dtype=dtype)
        field = field.reshape(grid.ActiveDimensions[::-1]).swapaxes(0,2)

        # we can/should also check against the max and min in the header file

        inFile.close()
        return field

    def _read_data_slice(self, grid, field, axis, coord):
        """wishful thinking?
        """
        sl = [slice(None), slice(None), slice(None)]
        sl[axis] = slice(coord, coord + 1)
        #sl = tuple(reversed(sl))
        return self._read_data_set(grid,field)[sl]


class IOHandlerPacked2D(IOHandlerPackedHDF5):

    _data_style = "enzo_packed_2d"
    _particle_reader = False

    def _read_data_set(self, grid, field):
        return HDF5LightReader.ReadData(grid.filename,
            "/Grid%08i/%s" % (grid.id, field)).transpose()[:,:,None]

    def modify(self, field):
        pass

    def _read_data_slice(self, grid, field, axis, coord):
        t = HDF5LightReader.ReadData(grid.filename, "/Grid%08i/%s" %
                        (grid.id, field)).transpose()
        return t


class IOHandlerPacked1D(IOHandlerPackedHDF5):

    _data_style = "enzo_packed_1d"
    _particle_reader = False

    def _read_data_set(self, grid, field):
        return HDF5LightReader.ReadData(grid.filename,
            "/Grid%08i/%s" % (grid.id, field)).transpose()[:,None,None]

    def modify(self, field):
        pass

    def _read_data_slice(self, grid, field, axis, coord):
        t = HDF5LightReader.ReadData(grid.filename, "/Grid%08i/%s" %
                        (grid.id, field))
        return t

class IOHandlerGadget(BaseIOHandler):
    _data_style = 'gadget_binary'
    def _read_data_set(self, grid, field):
        return grid._storage[field]

#
# Chombo readers
#

class IOHandlerChomboHDF5(BaseIOHandler):
    _data_style = "chombo_hdf5"
    _offset_string = 'data:offsets=0'
    _data_string = 'data:datatype=0'

    def _field_dict(self,fhandle):
        ncomp = int(fhandle['/'].attrs['num_components'])
        temp =  fhandle['/'].attrs.listitems()[-ncomp:]
        val, keys = zip(*temp)
        val = [int(re.match('component_(\d+)',v).groups()[0]) for v in val]
        return dict(zip(keys,val))
        
    def _read_field_names(self,grid):
        fhandle = h5py.File(grid.filename,'r')
        ncomp = int(fhandle['/'].attrs['num_components'])

        return [c[1] for c in f['/'].attrs.listitems()[-ncomp:]]
    
    def _read_data_set(self,grid,field):
        fhandle = h5py.File(grid.hierarchy.hierarchy_filename,'r')

        field_dict = self._field_dict(fhandle)
        lstring = 'level_%i' % grid.Level
        lev = fhandle[lstring]
        dims = grid.ActiveDimensions
        boxsize = dims.prod()
        
        grid_offset = lev[self._offset_string][grid._level_id]
        start = grid_offset+field_dict[field]*boxsize
        stop = start + boxsize
        data = lev[self._data_string][start:stop]

        return data.reshape(dims, order='F')
                                          

    def _read_data_slice(self, grid, field, axis, coord):
        sl = [slice(None), slice(None), slice(None)]
        sl[axis] = slice(coord, coord + 1)
        return self._read_data_set(grid,field)[sl]

class IOHandlerTiger(BaseIOHandler):
    _data_style = "tiger"
    _offset = 36

    def __init__(self, *args, **kwargs):
        BaseIOHandler.__init__(self, *args, **kwargs)
        self._memmaps = {}

    def _read_data_set(self, grid, field):
        fn = grid.pf.basename + grid.hierarchy.file_mapping[field]
        LD = na.array(grid.left_dims, dtype='int64')
        SS = na.array(grid.ActiveDimensions, dtype='int64')
        RS = na.array(grid.pf.root_size, dtype='int64')
        data = au.read_tiger_section(fn, LD, SS, RS).astype("float64")
        return data

    def _read_data_slice(self, grid, field, axis, coord):
        fn = grid.pf.basename + grid.hierarchy.file_mapping[field]
        LD = na.array(grid.left_dims, dtype='int64').copy()
        SS = na.array(grid.ActiveDimensions, dtype='int64').copy()
        RS = na.array(grid.pf.root_size, dtype='int64').copy()
        LD[axis] += coord
        SS[axis] = 1
        data = au.read_tiger_section(fn, LD, SS, RS).astype("float64")
        return data



"""
Import the components of the volume rendering extension

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as na
from yt.funcs import *
import h5py

from yt.amr_utils import PartitionedGrid, ProtoPrism, GridFace, \
        grid_points_in_volume, find_grids_in_inclined_box
from yt.lagos import ParallelAnalysisInterface, only_on_root, parallel_root_only
from yt.parallel_tools import DistributedObjectCollection

# HISTORICAL NOTE OF SOME IMPORT:
#   The homogenized brick class should and will be removed.  The more general,
#   and less fancy-parallel, HomogenizedVolume is the way to go.  The HBC is a
#   complicated mechanism that was designed for back when we were going to be
#   passing bricks all around between processors.

class HomogenizedVolume(ParallelAnalysisInterface):
    bricks = None
    def __init__(self, fields = "Density", source = None, pf = None,
                 log_fields = None):
        # Typically, initialized as hanging off a hierarchy.  But, not always.
        if pf is not None: self.pf = pf
        if source is None: source = self.pf.h.all_data()
        self.source = source
        self.fields = ensure_list(fields)
        if log_fields is not None:
            log_fields = ensure_list(log_fields)
        else:
            log_fields = [self.pf.field_info[field].take_log
                         for field in self.fields]
        self.log_fields = log_fields

    def traverse(self, back_point, front_point):
        if self.bricks is None: self.initialize_source()
        vec = front_point - back_point
        dist = na.minimum(
             na.sum((self.brick_left_edges - back_point) * vec, axis=1),
             na.sum((self.brick_right_edges - back_point) * vec, axis=1))
        ind = na.argsort(dist)
        for b in self.bricks[ind]: yield b

    def _partition_grid(self, grid):

        # This is not super efficient, as it re-fills the regions once for each
        # field.
        vcds = []
        for field, log_field in zip(self.fields, self.log_fields):
            vcd = grid.get_vertex_centered_data(field).astype('float64')
            if log_field: vcd = na.log10(vcd)
            vcds.append(vcd)

        GF = GridFaces(grid.Children + [grid])
        PP = ProtoPrism(grid.id, grid.LeftEdge, grid.RightEdge, GF)

        pgs = []
        for P in PP.sweep(0):
            sl = P.get_brick(grid.LeftEdge, grid.dds, grid.child_mask)
            if len(sl) == 0: continue
            dd = [d[sl[0][0]:sl[0][1]+1,
                    sl[1][0]:sl[1][1]+1,
                    sl[2][0]:sl[2][1]+1].copy() for d in vcds]
            pgs.append(PartitionedGrid(grid.id, len(self.fields), dd,
                        P.LeftEdge, P.RightEdge, sl[-1]))
        return pgs

    def initialize_source(self, source = None):
        if self.bricks is not None and source is not None:
            raise NotImplementedError("Sorry, dynamic shuffling of bricks is" + 
                                      " not yet supported")
        if self.bricks is not None and source is None: return
        bricks = []
        self._preload(self.source._grids, self.fields, self.pf.h.io)
        pbar = get_pbar("Partitioning ", len(self.source._grids))
        for i, g in enumerate(self.source._grids):
            pbar.update(i)
            bricks += self._partition_grid(g)
        pbar.finish()
        bricks = na.array(bricks, dtype='object')
        NB = len(bricks)
        # Now we set up our (local for now) hierarchy.  Note that to calculate
        # intersection, we only need to do the left edge & right edge.
        #
        # We're going to double up a little bit here in memory.
        self.brick_left_edges = na.zeros( (NB, 3), dtype='float64')
        self.brick_right_edges = na.zeros( (NB, 3), dtype='float64')
        self.brick_parents = na.zeros( NB, dtype='int64')
        self.brick_dimensions = na.zeros( (NB, 3), dtype='int64')
        for i,b in enumerate(bricks):
            self.brick_left_edges[i,:] = b.LeftEdge
            self.brick_right_edges[i,:] = b.RightEdge
            self.brick_parents[i] = b.parent_grid_id
            self.brick_dimensions[i,:] = b.my_data[0].shape
        # Vertex-centered means we subtract one from the shape
        self.brick_dimensions -= 1
        self.bricks = na.array(bricks, dtype='object')

    def store_bricks(self, fn):
        import h5py, cPickle
        f = h5py.File(fn, "w")
        f.create_dataset("/left_edges", data=self.brick_left_edges)
        f.create_dataset("/right_edges", data=self.brick_right_edges)
        f.create_dataset("/parents", data=self.brick_parents)
        f.create_dataset("/dimensions", data=self.brick_dimensions)
        f.create_group("/bricks")
        for i,b in enumerate(self.bricks):
            f.create_group("/bricks/brick_%08i" % i)
            for fi,field in enumerate(self.fields):
                f.create_dataset("/bricks/brick_%08i/%s" % (i, field),
                                 data=b.my_data[fi])
        f.close()

    def load_bricks(self, fn):
        import h5py
        f = h5py.File(fn, "r")
        self.brick_left_edges = f["/left_edges"][:]
        self.brick_right_edges = f["/right_edges"][:]
        self.brick_parents = f["/parents"][:]
        self.brick_dimensions= f["/dimensions"][:]
        bricks = []
        for i,ds in enumerate(sorted(f["/bricks"])):
            td = [f["/bricks/%s/%s" % (ds, field)][:] for field in self.fields]
            bricks.append(PartitionedGrid(
                                self.brick_parents[i], len(td), td,
                                self.brick_left_edges[i,:],
                                self.brick_right_edges[i,:],
                                self.brick_dimensions[i,:],
                                ))
        self.bricks = na.array(bricks, dtype='object')

class HomogenizedBrickCollection(DistributedObjectCollection):
    def __init__(self, source):
        # The idea here is that we have two sources -- the global_domain
        # source, which would be a decomposition of the 3D domain, and a
        # local_domain source, which is the set of bricks we want at the end.
        self.source = source
        self.pf = source.pf

    @classmethod
    def load_bricks(self, base_filename):
        pass

    def write_my_bricks(self, base_filename):
        pass

    def store_bricks(self, base_filename):
        pass
    
    @parallel_root_only
    def write_hierarchy(self, base_filename):
        pass
    
    def _partition_grid(self, grid, fields, log_field = None):
        fields = ensure_list(fields)
        if log_field is None: log_field = [True] * len(fields)

        # This is not super efficient, as it re-fills the regions once for each
        # field.
        vcds = []
        for i,field in enumerate(fields):
            vcd = grid.get_vertex_centered_data(field).astype('float64')
            if log_field[i]: vcd = na.log10(vcd)
            vcds.append(vcd)

        GF = GridFaces(grid.Children + [grid])
        PP = ProtoPrism(grid.id, grid.LeftEdge, grid.RightEdge, GF)

        pgs = []
        for P in PP.sweep(0):
            sl = P.get_brick(grid.LeftEdge, grid.dds, grid.child_mask)
            if len(sl) == 0: continue
            dd = [d[sl[0][0]:sl[0][1]+1,
                    sl[1][0]:sl[1][1]+1,
                    sl[2][0]:sl[2][1]+1].copy() for d in vcds]
            pgs.append(PartitionedGrid(grid.id, len(fields), dd,
                        P.LeftEdge, P.RightEdge, sl[-1]))
        return pgs

    def _partition_local_grids(self, fields = "Density", log_field = None):
        fields = ensure_list(fields)
        bricks = []
        # We preload.
        # UNCOMMENT FOR PARALLELISM
        #grid_list = list(self._get_grid_objs())
        grid_list = list(self.source._grids)
        self._preload(grid_list, fields, self.pf.h.io)
        pbar = get_pbar("Partitioning ", len(grid_list))
        # UNCOMMENT FOR PARALLELISM
        #for i, g in enumerate(self._get_grids()):
        print "THIS MANY GRIDS!", len(grid_list)
        for i, g in enumerate(self.source._grids):
            pbar.update(i)
            bricks += self._partition_grid(g, fields, log_field)
        pbar.finish()
        bricks = na.array(bricks, dtype='object')
        NB = len(bricks)
        # Now we set up our (local for now) hierarchy.  Note that to calculate
        # intersection, we only need to do the left edge & right edge.
        #
        # We're going to double up a little bit here in memory.
        self.brick_left_edges = na.zeros( (NB, 3), dtype='float64')
        self.brick_right_edges = na.zeros( (NB, 3), dtype='float64')
        self.brick_parents = na.zeros( NB, dtype='int64')
        self.brick_dimensions = na.zeros( (NB, 3), dtype='int64')
        self.brick_owners = na.ones(NB, dtype='int32') * self._mpi_get_rank()
        self._object_owners = self.brick_owners
        for i,b in enumerate(bricks):
            self.brick_left_edges[i,:] = b.LeftEdge
            self.brick_right_edges[i,:] = b.RightEdge
            self.brick_parents[i] = b.parent_grid_id
            self.brick_dimensions[i,:] = b.my_data[0].shape
        # Vertex-centered means we subtract one from the shape
        self.brick_dimensions -= 1
        self.bricks = na.array(bricks, dtype='object')
        # UNCOMMENT FOR PARALLELISM
        #self.join_lists()

    def _get_object_info(self):
        # We transpose here for the catdict operation
        info_dict = dict(left_edges = self.brick_left_edges.transpose(),
                         right_edges = self.brick_right_edges.transpose(),
                         parents = self.brick_parents,
                         owners = self.brick_owners,
                         dimensions = self.brick_dimensions.transpose(),)
        return info_dict

    def _set_object_info(self, info_dict):
        self.brick_left_edges = info_dict.pop("left_edges").transpose()
        self.brick_right_edges = info_dict.pop("right_edges").transpose()
        self.brick_parents = info_dict.pop("parents")
        self.brick_dimensions = info_dict.pop("dimensions").transpose()
        self.brick_owners = info_dict.pop("owners")
        self._object_owners = self.brick_owners
        bricks = self.bricks
        self.bricks = na.array([None] * self.brick_owners.size, dtype='object')
        # Copy our bricks back in
        self.bricks[self.brick_owners == self._mpi_get_rank()] = bricks[:]

    def _create_buffer(self, ind_list):
        # Note that we have vertex-centered data, so we add one before taking
        # the prod and the sum
        total_size = (self.brick_dimensions[ind_list,:] + 1).prod(axis=1).sum()
        mylog.debug("Creating buffer for %s bricks (%s)",
                    len(ind_list), total_size)
        my_buffer = na.zeros(total_size, dtype='float64')
        return my_buffer

    def _pack_buffer(self, ind_list, my_buffer):
        si = 0
        for index in ind_list:
            d = self.bricks[index].my_data.ravel()
            my_buffer[si:si+d.size] = d[:]
            si += d.size

    def _unpack_buffer(self, ind_list, my_buffer):
        si = 0
        for index in ind_list:
            pgi = self.brick_parents[index]
            LE = self.brick_left_edges[index,:].copy()
            RE = self.brick_right_edges[index,:].copy()
            dims = self.brick_dimensions[index,:].copy()
            size = (dims + 1).prod()
            data = my_buffer[si:si+size].reshape(dims + 1)
            self.bricks[index] = PartitionedGrid(
                    pgi, data, LE, RE, dims)
            si += size

    def _wipe_objects(self, indices):
        self.bricks[indices] = None

    def _collect_bricks(self, intersection_source):
        if not self._distributed: return
        # This entire routine should instead be set up to do:
        #   alltoall broadcast of the *number* of requested bricks
        #   non-blocking receives posted for int arrays
        #   sizes of data calculated
        #   concatenated data receives posted
        #   send all data
        #   get bricks back
        # This presupposes that we are using the AMRInclinedBox as a data
        # source.  If we're not, we ought to be.
        needed_brick_i = find_grids_in_inclined_box(
            intersection_source.box_vectors, intersection_source.center,
            self.brick_left_edges, self.brick_right_edges)
        needed_brick_i = na.where(needed_brick_i)[0]
        self._collect_objects(needed_brick_i)

    def _initialize_parallel(self):
        pass

    def _finalize_parallel(self):
        pass

    def get_brick(self, brick_id):
        pass

    @property
    def _grids(self):
        return self.source._grids

class GridFaces(object):
    def __init__(self, grids):
        self.faces = [ [], [], [] ]
        for grid in grids:
            for direction in range(3):
                self.faces[direction].append( GridFace(grid, direction, 1) )
                self.faces[direction].append( GridFace(grid, direction, 0) )
        for f in self.faces:
            f.sort(key = lambda a: a.coord)

    def __getitem__(self, item):
        return self.faces[item]

def export_partitioned_grids(grid_list, fn, int_type=na.int64, float_type=na.float64):
    f = h5py.File(fn, "w")
    pbar = get_pbar("Writing Grids", len(grid_list))
    nelem = sum((grid.my_data.size for grid in grid_list))
    ngrids = len(grid_list)
    group = f.create_group("/PGrids")
    left_edge = na.concatenate([[grid.LeftEdge,] for grid in grid_list])
    f.create_dataset("/PGrids/LeftEdges", data=left_edge, dtype=float_type); del left_edge
    right_edge = na.concatenate([[grid.RightEdge,] for grid in grid_list])
    f.create_dataset("/PGrids/RightEdges", data=right_edge, dtype=float_type); del right_edge
    dims = na.concatenate([[grid.my_data.shape[:],] for grid in grid_list])
    f.create_dataset("/PGrids/Dims", data=dims, dtype=int_type); del dims
    data = na.concatenate([grid.my_data.ravel() for grid in grid_list])
    f.create_dataset("/PGrids/Data", data=data, dtype=float_type); del data
    f.close()
    pbar.finish()

def import_partitioned_grids(fn, int_type=na.int64, float_type=na.float64):
    f = h5py.File(fn, "r")
    n_groups = len(f.listnames())
    grid_list = []
    dims = f["/PGrids/Dims"][:].astype(int_type)
    left_edges = f["/PGrids/LeftEdges"][:].astype(float_type)
    right_edges = f["/PGrids/RightEdges"][:].astype(float_type)
    data = f["/PGrids/Data"][:].astype(float_type)
    pbar = get_pbar("Reading Grids", dims.shape[0])
    curpos = 0
    for i in xrange(dims.shape[0]):
        gd = dims[i,:]
        gle, gre = left_edges[i,:], right_edges[i,:]
        gdata = data[curpos:curpos+gd.prod()].reshape(gd)
        # Vertex -> Grid, so we -1 from dims in this
        grid_list.append(PartitionedGrid(-1, gdata, gle, gre, gd - 1))
        curpos += gd.prod()
        pbar.update(i)
    pbar.finish()
    f.close()
    return na.array(grid_list, dtype='object')



from yt.config import ytcfg

ytcfg["yt","time_functions"] = "True"

from yt.mods import *

yt_counters("Full Time")

yt_counters("yt Hierarchy")
pf = load('data0005')

pf.h
yt_counters("yt Hierarchy")

h = yt.lagos.HaloFinding.parallelHF(pf, threshold=160.0, safety=2.5, \
dm_only=False,resize=True, fancy_padding=True, rearrange=True)

yt_counters("Writing Data")
h.write_out('dist-chain.out')
h.write_particle_lists_txt("chain")
h.write_particle_lists("chain")
yt_counters("Writing Data")

yt_counters("Full Time")



"""
Simle integrators for the radiative transfer equation

Author: Britton Smith <brittonsmith@gmail.com>
Affiliation: CASA/University of Colorado
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2008 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

@cython.boundscheck(False)
@cython.wraparound(False)
def CICDeposit_3(np.ndarray[np.float64_t, ndim=1] posx,
                 np.ndarray[np.float64_t, ndim=1] posy,
                 np.ndarray[np.float64_t, ndim=1] posz,
                 np.ndarray[np.float32_t, ndim=1] mass,
                 np.int64_t npositions,
                 np.ndarray[np.float32_t, ndim=3] field,
                 np.ndarray[np.float64_t, ndim=1] leftEdge,
                 np.ndarray[np.int32_t, ndim=1] gridDimension,
                 np.float64_t cellSize):

    cdef int i1, j1, k1, n
    cdef double xpos, ypos, zpos
    cdef double fact, edge0, edge1, edge2
    cdef double le0, le1, le2
    cdef float dx, dy, dz, dx2, dy2, dz2

    edge0 = (<float> gridDimension[0]) - 0.5001
    edge1 = (<float> gridDimension[1]) - 0.5001
    edge2 = (<float> gridDimension[2]) - 0.5001
    fact = 1.0 / cellSize

    le0 = leftEdge[0]
    le1 = leftEdge[1]
    le2 = leftEdge[2]

    for n in range(npositions):

        # Compute the position of the central cell
        xpos = fmin(fmax((posx[n] - le0)*fact, 0.5001), edge0)
        ypos = fmin(fmax((posy[n] - le1)*fact, 0.5001), edge1)
        zpos = fmin(fmax((posz[n] - le2)*fact, 0.5001), edge2)

        i1  = <int> (xpos + 0.5)
        j1  = <int> (ypos + 0.5)
        k1  = <int> (zpos + 0.5)

        # Compute the weights
        dx = (<float> i1) + 0.5 - xpos
        dy = (<float> j1) + 0.5 - ypos
        dz = (<float> k1) + 0.5 - zpos
        dx2 =  1.0 - dx
        dy2 =  1.0 - dy
        dz2 =  1.0 - dz

        # Interpolate from field into sumfield
        field[i1-1,j1-1,k1-1] += mass[n] * dx  * dy  * dz
        field[i1  ,j1-1,k1-1] += mass[n] * dx2 * dy  * dz
        field[i1-1,j1  ,k1-1] += mass[n] * dx  * dy2 * dz
        field[i1  ,j1  ,k1-1] += mass[n] * dx2 * dy2 * dz
        field[i1-1,j1-1,k1  ] += mass[n] * dx  * dy  * dz2
        field[i1  ,j1-1,k1  ] += mass[n] * dx2 * dy  * dz2
        field[i1-1,j1  ,k1  ] += mass[n] * dx  * dy2 * dz2
        field[i1  ,j1  ,k1  ] += mass[n] * dx2 * dy2 * dz2



"""
Main application for Reason.  Includes the basic window outline.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2007-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


from yt.reason import *

_StaticOutputMenuItems = ["proj","slice","export",]
_SphereObjectMenuItems = ["phase","profile","cutting","export","extract"]
_ProjObjectMenuItems = ["export",]
_SliceObjectMenuItems = []
_CuttingPlaneObjectMenuItems = ["export",]
_ProfileObjectMenuItems = ["export",]
_FidoOutputMenuItems = ["delete",]

class ReasonMainWindow(wx.Frame):
    def __init__(self, *args, **kwds):
        kwds["style"] = wx.DEFAULT_FRAME_STYLE
        kwds["title"] = "yt - Reason"
        kwds["size"] = (ytcfg.getint("reason","width"),
                        ytcfg.getint("reason","height"))
        ytcfg.set('yt','inGui','True')
        wx.Frame.__init__(self, *args, **kwds)

        self.__setup_controls()

        self.__setup_menubar()
        self.__setup_popup_menu()
        self.__setup_toolbar()
        self.__setup_data_tree()

        self.status_bar = self.CreateStatusBar(4, 0)

        self.__set_properties()
        self.__do_layout()

        Publisher().subscribe(self.MessagePageDeleted, ('page_deleted'))
        self.Bind(wx.EVT_CLOSE, self.OnClose)

    def __setup_controls(self):

        self.main_splitter = wx.SplitterWindow(self, -1)
        self.view_panel = wx.SplitterWindow(self.main_splitter, -1)
        self.data_panel = wx.Panel(self.view_panel, -1)
        self.data_tree = wx.TreeCtrl(self.data_panel, -1,
                style=wx.TR_HIDE_ROOT | wx.TR_LINES_AT_ROOT | wx.TR_HAS_BUTTONS)
        self.plot_panel = PlotPanel(parent=self.view_panel)
        self.__setup_interpreter()

        self.main_splitter.SetMinimumPaneSize(20)
        self.main_splitter.SplitHorizontally(self.view_panel, self.int_panel, -100)
        self.view_panel.SetMinimumPaneSize(20)
        self.view_panel.SplitVertically(self.data_panel, self.plot_panel, 200)

    def __setup_interpreter(self):
        self.windows = []
        self.outputs = []
        self.data_objects = []
        self.locals = {'lagos':lagos,
                       'raven':raven,
                       'raven':raven,
                       'outputs':self.outputs,
                       'windows':self.windows,
                       'mainwindow':self,
                       'data_objects':self.data_objects,
                       'pylab':pylab,
                       'add_profile':self.__add_profile_wrapper,
                       'add_phase':self.__add_phase_wrapper}
        wx.py.buffer.Buffer.updateNamespace = \
                get_new_updateNamespace(self.locals)
        self.int_panel = wx.Panel(self.main_splitter, -1)
        #self.interpreter = ReasonInterpreterPanel(self.int_panel, -1, self.locals)
        self.interpreter = LoggingWindowBox(self.int_panel, -1)

    def __setup_menubar(self):
        menu_bar = wx.MenuBar()
        file_menu = wx.Menu()
        menu_bar.Append(file_menu, "File")

        # Set up IDs for event binding

        open_hierarchy = file_menu.Append(-1, "Open Individual Hierarchy")
        import_directory = file_menu.Append(-1, "Import Data Directory")
        file_menu.AppendSeparator()
        field_inspector = file_menu.Append(-1, "Inspect Fields")
        #open_shell = file_menu.Append(-1, "Open Shell")
        open_editor = file_menu.Append(-1, "Open Shell & Editor")
        save_image = file_menu.Append(-1, "Save Image")
        file_menu.AppendSeparator()
        exit = file_menu.Append(-1, "Exit")

        self.Bind(wx.EVT_MENU, self.OnOpenHierarchy, open_hierarchy)
        self.Bind(wx.EVT_MENU, self.OnImportDirectory, import_directory)
        self.Bind(wx.EVT_MENU, self.OnInspectFields, field_inspector)
        #self.Bind(wx.EVT_MENU, self.OnOpenShell, open_shell)
        self.Bind(wx.EVT_MENU, self.OnOpenEditor, open_editor)
        self.Bind(wx.EVT_MENU, self.OnSaveImage, save_image)
        self.Bind(wx.EVT_MENU, self.OnExit, exit)

        self.SetMenuBar(menu_bar)

    def __setup_popup_menu(self):
        self.PopupMenu = wx.Menu()
        self.PopupMenuIds = {}
        self.PopupMenuIds["slice"] = self.PopupMenu.Append(-1, "Slice")
        self.PopupMenuIds["proj"] = self.PopupMenu.Append(-1, "Project")
        self.PopupMenuIds["phase"] = self.PopupMenu.Append(-1, "Phase Plot")
        self.PopupMenuIds["profile"] = self.PopupMenu.Append(-1, "Profile Plot")
        self.PopupMenuIds["cutting"] = self.PopupMenu.Append(-1, "Cutting Plane")
        self.PopupMenuIds["extract"] = self.PopupMenu.Append(-1, "Extract Set")
        self.PopupMenuIds["cube"] = self.PopupMenu.Append(-1, "Extract Cube")
        self.PopupMenuIds["export"] = self.PopupMenu.Append(-1, "Export Data")
        self.PopupMenuIds["delete"] = self.PopupMenu.Append(-1, "Delete Item")

        self.Bind(wx.EVT_MENU, self._add_slice, self.PopupMenuIds["slice"])
        self.Bind(wx.EVT_MENU, self._add_proj, self.PopupMenuIds["proj"])
        self.Bind(wx.EVT_MENU, self._add_phase, self.PopupMenuIds["phase"])
        self.Bind(wx.EVT_MENU, self._add_profile, self.PopupMenuIds["profile"])
        self.Bind(wx.EVT_MENU, self._add_cutting, self.PopupMenuIds["cutting"])
        self.Bind(wx.EVT_MENU, self._extract_set, self.PopupMenuIds["extract"])
        #self.Bind(wx.EVT_MENU, self._extract_cube, self.PopupMenuIds["cube"])
        self.Bind(wx.EVT_MENU, self._export_data_object, self.PopupMenuIds["export"])
        self.Bind(wx.EVT_MENU, self._remove_fido_output, self.PopupMenuIds["delete"])

    def __setup_data_tree(self):

        self.root = self.data_tree.AddRoot("You shouldn't see me!")
        self.fido_root = self.data_tree.AppendItem(self.root, "Stored Outputs")
        self.output_root = self.data_tree.AppendItem(self.root, "Selected Outputs")
        self.data_root = self.data_tree.AppendItem(self.root, "Data Objects")

        self.data_tree.Expand(self.fido_root)
        self.data_tree.Expand(self.output_root)
        self.data_tree.Expand(self.data_root)

        self.data_tree.Bind(wx.EVT_RIGHT_DOWN, self.OnRightDown)
        self.data_tree.Bind(wx.EVT_RIGHT_DOWN, self.OnRightDown)
        self.data_tree.Bind(wx.EVT_LEFT_DCLICK, self.OnRightDown)
        self.Bind(wx.EVT_TREE_ITEM_EXPANDED, self.OnItemExpanded, self.data_tree)

        self.__setup_fido_tree()

    def __setup_fido_tree(self, event=None):
        # Calling this delete may not be wise.
        # However, as long as we have a distinction between
        # the data outputs and the created data objects, it should be okay.
        self.data_tree.DeleteChildren(self.fido_root)
        gc = fido.GrabCollections()
        for c in gc:
            tid = wx.TreeItemData((c.title, c._last_read_filename,
                                  -1, _FidoOutputMenuItems,))
            cRoot = self.data_tree.AppendItem(self.fido_root, c.title, data=tid)
            for fn in c:
                if not os.path.isfile(fn): continue
                try:
                    z = str(fido.get_parameter_line(fn,
                                 "CosmologyCurrentRedshift"))
                except:
                    z = "N/A"
                tt = str(fido.get_parameter_line(fn,
                             "InitialTime"))
                tid = wx.TreeItemData((fn, tt, z, _StaticOutputMenuItems, ))
                ni = self.data_tree.AppendItem(cRoot,
                    "%s" % (os.path.basename(fn)), data=tid)

    def _remove_fido_output(self, event=None):
        tid = self.data_tree.GetSelection()
        data_object = self.data_tree.GetItemData(tid).GetData()
        dlg = wx.MessageDialog(self,
                'Are you sure you want to delete the record of that output?\n' + \
                'This will leave the data itself unaffected.',
                'Record Deletion Confirmation', wx.YES_NO | wx.ICON_INFORMATION)
        if dlg.ShowModal() == wx.ID_YES and os.path.isfile(data_object[1]):
            mylog.info("Removing %s", data_object[1])
            os.unlink(data_object[1])
        dlg.Destroy()
        self.__setup_fido_tree()

    def __setup_toolbar(self):
        # Tool Bar
        self._VMTB_REREADFIDO = wx.NewId()
        self._VMTB_FULLDOMAIN = wx.NewId()
        self._VMTB_CHANGEZOOM = wx.NewId()
        self._VMTB_REDRAW = wx.NewId()
        self._VMTB_SAVE = wx.NewId()
        self._VMTB_FIELDSWITCH = wx.NewId()
        self._VMTB_CHANGELIMITS = wx.NewId()
        self._VMTB_VIEWPF = wx.NewId()
        self._VMTB_VELPLOT = wx.NewId()

        self.toolbar = wx.ToolBar(self, -1, style=wx.TB_HORIZONTAL|wx.TB_TEXT|wx.TB_HORZ_LAYOUT)
        font = self.toolbar.GetFont()
        font.SetFamily(wx.MODERN)
        self.toolbar.SetFont(font)
        self.toolbar.SetToolBitmapSize((16,16))

        self.SetToolBar(self.toolbar)
        def AddButton(id, label, tooltip="", bitmapID=None):
            if bitmapID != None:
                bm = wx.ArtProvider.GetBitmap(bitmapID, wx.ART_TOOLBAR, (16,16))
            else: bm = wx.NullBitmap
            self.toolbar.AddLabelTool(id, label, bm, bm, wx.ITEM_NORMAL, tooltip, "")
            self.toolbar.AddSeparator()

        self.toolbar.AddSeparator()
        AddButton(self._VMTB_REREADFIDO,"Update OutputList",
                                        "Reread from the Fido database", wx.ART_TIP)
        AddButton(self._VMTB_REDRAW,"Redraw", "Force a redraw", wx.ART_REDO)
        self.available_fields = wx.Choice(self.toolbar, id=self._VMTB_FIELDSWITCH, choices = [])
        self.toolbar.AddControl(self.available_fields)
        Publisher().subscribe(self.MessageUpdateToolbarFields, ('page_changed'))
        AddButton(self._VMTB_FULLDOMAIN, "Zoom Top",  "Zoom to the top level", wx.ART_FIND)
        AddButton(self._VMTB_CHANGELIMITS, "Change Limits", "Change the colorbar limits", wx.ART_GO_UP)
        AddButton(self._VMTB_VIEWPF, "View ParameterFile", "View the parameter file", wx.ART_NORMAL_FILE)
        cl = wx.ArtProvider.GetBitmap(wx.ART_TICK_MARK, wx.ART_TOOLBAR, (16,16))
        #self.toolbar.AddCheckLabelTool(self._VMTB_VELPLOT, "VelVecs", cl, shortHelp="Plot Velocity Vectors")
        self.toolbar.AddSeparator()

        self.Bind(wx.EVT_MENU, self.__setup_fido_tree, id=self._VMTB_REREADFIDO)
        self.Bind(wx.EVT_CHOICE, self.plot_panel.OnCallSwitchField, id=self._VMTB_FIELDSWITCH)
        self.Bind(wx.EVT_MENU, self.plot_panel.OnCallSetWidth, id=self._VMTB_CHANGEZOOM)
        self.Bind(wx.EVT_MENU, self.plot_panel.OnCallRedraw, id=self._VMTB_REDRAW)
        self.Bind(wx.EVT_MENU, self.plot_panel.OnCallZoomTop, id=self._VMTB_FULLDOMAIN)
        self.Bind(wx.EVT_MENU, self.plot_panel.OnCallSetZLim, id=self._VMTB_CHANGELIMITS)
        self.Bind(wx.EVT_MENU, self.plot_panel.OnCallViewPF, id=self._VMTB_VIEWPF)

    def __set_properties(self):
        self.toolbar.SetToolBitmapSize((24, 24))
        self.toolbar.Realize()
        self.status_bar.SetStatusWidths([-1,-1,-1,-1])

    def __do_layout(self):
        MainWindowSizer = wx.BoxSizer(wx.VERTICAL)
        DataPanelSizer = wx.BoxSizer(wx.VERTICAL)

        IntPanelSizer = wx.BoxSizer(wx.HORIZONTAL)
        IntPanelSizer.Add(self.interpreter, 1, wx.EXPAND, 0)
        self.int_panel.SetSizer(IntPanelSizer)

        DataPanelSizer.Add(self.data_tree, 1, wx.EXPAND, 0)
        self.data_panel.SetSizer(DataPanelSizer)
        self.data_panel.Layout()

        MainWindowSizer.Add(self.main_splitter, 1, wx.EXPAND)
        self.SetSizer(MainWindowSizer)

        self.Layout()

    def _update_toolbar_fields(self, page):
        newItems = None
        if hasattr(page,'QueryFields'): newItems = page.QueryFields()
        if not newItems:
            self.available_fields.Enable(False)
        else:
            self.available_fields.Enable(True)
            self.available_fields.SetItems(newItems)

    def _add_static_output(self, filename):
        # Alright, we choose the hierarchy in the file selector,
        # so let's strip that extension off
        fn = filename[:-10]
        eso = lagos.EnzoStaticOutput(fn)
        try:
            z = str(eso["CosmologyCurrentRedshift"])
        except:
            z = "N/A"
        mylog.info("Adding output %s", fn)
        tid = wx.TreeItemData((eso, str(eso["InitialTime"]), z, _StaticOutputMenuItems))
        self.outputs.append(eso)
        ni = self.data_tree.AppendItem(self.output_root, "%s" % (eso.basename), data=tid)
        self.data_tree.Expand(self.output_root)

    def _add_data_object(self, title, object, mids, parent_id = None):
        self.data_objects.append(object)
        tid = wx.TreeItemData((object, title, len(self.data_objects), mids))
        if parent_id is None: parent_id = self.data_root
        ni = self.data_tree.AppendItem(parent_id, "%s" % (title), data=tid)
        self.data_tree.Expand(parent_id)

    def _add_sphere(self, title, sphere, parent_id=None):
        # These all get passed in
        self._add_data_object(title, sphere, _SphereObjectMenuItems, parent_id)

    def __add_profile_wrapper(self, obj):
        """
        Add a profile plot from an arbitrary object.
        """
        pass

    def __add_phase_wrapper(self, obj):
        """
        Add a phase plot from an arbitrary object.
        """
        self._add_phase(data_object=obj)

    def _add_page_to_notebook(self, page, name, id):
        self.windows.append(page)
        self.plot_panel.AddPlot(self.windows[-1], name, id)
        mylog.debug("Adding page with ID: %s", id)
        wx.SafeYield(onlyIfNeeded = True)
        self.plot_panel.UpdateSubscriptions(page)

    def _add_profile(self, event=None, data_object = None):
        MyID = wx.NewId()
        parent_id = None
        if data_object is None: parent_id, data_object = self.get_output()
        p1ds = Profile1DSetup(data_object, self)
        if not p1ds.ShowModal() == wx.ID_OK:
            p1ds.Destroy()
            return
        argdict = p1ds.return_argdict()
        if argdict is None:
            p1ds.Destroy()
            return
        if argdict['lower_bound']is None:
           argdict['lower_bound']= \
                self.__find_min(data_object, argdict['bin_field'])
        if argdict['upper_bound'] is None:
           argdict['upper_bound'] = \
                self.__find_max(data_object, argdict['bin_field'])
        self._add_page_to_notebook(
            Profile1DPlotPage(parent=self.plot_panel.nb,
                            status_bar=self.status_bar,
                            data_object = data_object,
                            argdict = argdict, CreationID = MyID,
                            mw = self, parent_id=parent_id),
            "Profile Plot %s" % MyID, MyID)
        if parent_id is not None:
            self._add_data_object("Profile: %s" % (argdict['bin_field']),
                               self.windows[-1].plot.data,
                               _ProfileObjectMenuItems, parent_id)


    def _add_phase(self, event=None, data_object = None):
        MyID = wx.NewId()
        parent_id = None
        if data_object is None: parent_id, data_object = self.get_output()
        p2ds = Profile2DSetup(data_object, self)
        if not p2ds.ShowModal() == wx.ID_OK:
            p2ds.Destroy()
            return
        argdict = p2ds.return_argdict()
        if argdict is None:
            p2ds.Destroy()
            return
        for ax in 'xy':
            if argdict['%s_lower_bound'%ax] is None:
                argdict['%s_lower_bound'%ax] = \
                    self.__find_min(data_object, argdict['%s_bin_field'%ax])
            if argdict['%s_upper_bound'%ax] is None:
                argdict['%s_upper_bound'%ax] = \
                    self.__find_max(data_object, argdict['%s_bin_field'%ax])
        self._add_page_to_notebook(
            PhasePlotPage(parent=self.plot_panel.nb,
                          status_bar=self.status_bar,
                          data_object = data_object,
                          argdict = argdict, CreationID = MyID,
                          mw = self, parent_id=parent_id),
            "Phase Plot %s" % MyID, MyID)
        if parent_id is not None:
            self._add_data_object("Phase: %s, %s" % (argdict['x_bin_field'],
                                                 argdict['y_bin_field']),
                               self.windows[-1].plot.data,
                               _ProfileObjectMenuItems, parent_id)

    def __find_min(self, data_object, field):
        return data_object[field].min()

    def __find_max(self, data_object, field):
        return data_object[field].max()

    def _add_proj(self, event=None):
        MyID = wx.NewId()
        parent_id, data_object = self.get_output()
        width = 1.0
        unit = "1"
        proj_setup = ProjectionSetup(data_object, self)
        if not proj_setup.ShowModal() == wx.ID_OK:
            proj_setup.Destroy()
            return
        field = proj_setup.field.GetStringSelection()
        weight_field = proj_setup.weight_field.GetStringSelection()
        if weight_field == "": weight_field = None
        total = 0
        for i, ax in enumerate('xyz'):
            if not getattr(proj_setup,'%s_ax' % ax).GetValue(): continue
            mylog.info("Adding %s projection of %s" % (ax, data_object))
            self._add_page_to_notebook(
                ProjPlotPage(parent=self.plot_panel.nb,
                              status_bar=self.status_bar,
                              outputfile = data_object,
                              axis=i,
                              field = field,
                              weight_field = weight_field,
                              mw = self, CreationID=MyID,
                              parent_id=parent_id),
                "%s - Projection - %s" % (data_object.basename, ax),
                MyID)
            self._add_data_object("Proj: %s - %s (%s)" % (ax, field, weight_field),
                               self.windows[-1].plot.data,
                               _ProjObjectMenuItems, parent_id)
            print "Adding with ID:", MyID
            total += 1
        for w in self.windows[-total:]: w.ChangeWidth(1,'1')
        proj_setup.Destroy()

    def _add_slice(self, event=None):
        MyID = wx.NewId()
        parent_id, data_object = self.get_output()
        field, width, unit = "Density", 1.0, '1'
        for i, ax in enumerate('xyz'):
            mylog.info("Adding %s slice of %s" % (ax, data_object))
            self._add_page_to_notebook(
                SlicePlotPage(parent=self.plot_panel.nb,
                              status_bar=self.status_bar,
                              outputfile = data_object,
                              axis=i,
                              field = field,
                              mw = self, CreationID=MyID,
                              parent_id = parent_id),
                "%s - Slice - %s" % (data_object.basename, ax),
                MyID)
            self._add_data_object("Slice: %s" % (ax),
                               self.windows[-1].plot.data,
                               _SliceObjectMenuItems, parent_id)
        for w in self.windows[-3:]: w.ChangeWidth(1,'1')

    def _export_data_object(self, event):
        parent_id, data_object = self.get_output()
        # Get a file name and then attempt to export.
        dlg = wx.FileDialog( \
            self, message="Save Data As ...", defaultDir=os.getcwd(), \
            defaultFile="exported.dat", wildcard="dat (*.dat)|*.dat", style=wx.SAVE)
        if dlg.ShowModal() == wx.ID_OK:
            try:
                path = dlg.GetPath()
                data_object.write_out(path)
            except (ValueError, AttributeError), exc:
                print exc, dir(exc)
                err = wx.MessageDialog(None, "Error",
                                "The writing out has failed.\n\n" + \
                                "See the console for more information.")
                err.ShowModal()
                err.Destroy()
        dlg.Destroy()

    def _extract_set(self, event=None, data_object=None):
        MyID = wx.NewId()
        parent_id = None
        if data_object is None: parent_id, data_object = self.get_output()
        ess = ExtractSetSetup(data_object, self)
        if not ess.ShowModal() == wx.ID_OK:
            ess.Destroy()
            return
        ind = ess.return_indices()
        ess.Destroy()
        extracted_set = data_object.extract_region(ind)
        self._add_data_object("Extracted Set",
                              extracted_set,
                              _SphereObjectMenuItems, parent_id)

    def __add_cutting_wrapper(self, parameter_file, normal):
        self._add_cutting(parameter_file=parameter_file, normal=normal)

    def _add_cutting(self, event=None, parameter_file = None, normal=None,
                     center = None):
        parent_id = None
        if parameter_file is None or normal is None or center is None:
            parent_id, data_object = self.get_output()
            data_object.set_field_parameter("bulk_velocity",
                data_object.quantities["BulkVelocity"](lazy_reader=True))
            normal = data_object.quantities["AngularMomentumVector"](lazy_reader=True)
            center = data_object.get_field_parameter("center")
            parameter_file = data_object.pf
        MyID = wx.NewId()
        field, width, unit = "Density", 1.0, '1'
        mylog.info("Adding cutting plane of %s with normal %s",
                   data_object, normal)
        self._add_page_to_notebook(
            CuttingPlanePlotPage(parent=self.plot_panel.nb,
                            status_bar=self.status_bar,
                            outputfile=parameter_file, field=field, mw=self,
                            CreationID=MyID, axis=4, normal=normal,
                            center = center, parent_id=parent_id),
            "%s - Cutting Plane" % (parameter_file.basename), MyID)
        self._add_data_object("Cutting Plane" % (parameter_file),
                              self.windows[-1].plot.data,
                              _CuttingPlaneObjectMenuItems, parent_id)
        self.windows[-1].ChangeWidth(1,'1')

    def get_output(self, event=None):
        # Figure out which outputs are selected
        #tid = self.data_tree.GetFirstSelected()
        tid = self.data_tree.GetSelection()
        ii = self.data_tree.GetItemData(tid).GetData()[0]
        if isinstance(ii, types.StringTypes): # We promote
            ii = lagos.EnzoStaticOutput(ii) # Instantiate here
            self.outputs.append(ii)
            fn, z, t, mids = self.data_tree.GetItemData(tid).GetData()
            newData = wx.TreeItemData((ii, z, t, mids))
            self.data_tree.SetItemData(tid, newData)
        print "Got output:", ii
        return tid, ii

    # Functions bound to messages in pubsub

    def MessageUpdateToolbarFields(self, message):
        page = message.data
        self._update_toolbar_fields(page)

    def MessagePageDeleted(self, message):
        id = message.data
        del self.windows[id]

    # Functions bound exclusively to events:

    def OnExit(self, event):
        self.Close()

    def OnClose(self, event):
        Publisher().unsubAll()
        self.Destroy()

    def OnInspectFields(self, event):
        p = FieldFunctionInspector(self, -1)
        p.Show()

    def OnImportDirectory(self, event=None):
        dialog = wx.DirDialog(self, "Choose a data directory",
                    style=wx.DD_DEFAULT_STYLE | wx.DD_DIR_MUST_EXIST)
        if dialog.ShowModal() == wx.ID_OK:
            file = dialog.GetPath()
            fido.Import(current_path = file).PerformAction()
        self.__setup_fido_tree()
        dialog.Destroy()

    def OnOpenHierarchy(self, event):
        wildcard = "Hierarchy (*.hierarchy)|*.hierarchy|" \
                   "All files (*,*)|*.*"
        dialog = wx.FileDialog(None, "Choose your hierarchy", os.getcwd(),
                               "", wildcard, wx.OPEN)
        if dialog.ShowModal() == wx.ID_OK:
            file = dialog.GetPath()
            self._add_static_output(file)
        dialog.Destroy()

    def OnOpenEditor(self, event):
        frame = ReasonEditorNotebookFrame(parent=self,
                                          title="Editor")
        frame.SetStatusText("Reason Shell")
        frame.Show()
        self.ff = frame

    def OnOpenShell(self, event):
        frame = wx.py.shell.ShellFrame(parent=self, locals=self.locals)
        frame.SetStatusText("Reason Shell")
        frame.Show()

    def OnSaveImage(self, event):
        pgI = self.plot_panel.nb.Selection
        pg = self.plot_panel.nb.GetPage(pgI)
        if not hasattr(pg, 'SaveImage'): return
        pg.SaveImage()

    def OnItemExpanded(self, event):
        if event.GetItem() == self.fido_root:
            mylog.info("Reloading fido outputs")
            self.__setup_fido_tree()

    def OnRightDown(self, event):
        pt = event.GetPosition();
        item, flags = self.data_tree.HitTest(pt)
        if item:
            self.data_tree.SelectItem(item)
            pos = event.GetPosition()
            self.ContextMenuPosition = pos
            itemData = self.data_tree.GetItemData(item).Data
            if not itemData: return
            for n,d in self.PopupMenuIds.items():
                self.PopupMenu.Enable(d.Id,False)
            if itemData:
                for n in itemData[3]:
                    self.PopupMenu.Enable(self.PopupMenuIds[n].Id, True)
            #self.PopupMenu.Enable(self.PopupMenuIds["proj"].Id,False)
            self.data_tree.PopupMenu(self.PopupMenu, pos)



"""
Notebook pages and main plotpanel classes.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2007-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


from yt.reason import *

_WelcomeMessage = \
"""
Welcome to Reason.

This GUI was designed as a front-end to a pre-existing toolkit called yt.  As such, it is feature-incomplete -- there's so much more that you can do with the command line.  That aside, there's some fun stuff you can do just in here.

To access the various plotting functions, right click on a dataset in the tree over to the left.  Then you can right-click in the window to shuffle around the center, to change the color map, to save it, and so on.  And then give a shot to shift-clicking in two places -- the first will define the center, the second will define the outer-edge.  Then you can phase plot the sphere!

If you experience any troubles, drop me a line (matthewturk@gmail.com) or just fill out a ticket at yt.enzotools.org.
"""

class PlotPanel(wx.Panel):
    def __init__(self, *args, **kwds):
        # begin wxGlade: ReasonVMPlotFrame.__init__
        self.parent = kwds["parent"]
        self.CurrentlyResizing = False
        wx.Panel.__init__(self, *args, **kwds)
        self._link_plots = True

        self._AUI_NOTEBOOK = wx.NewId()
        self.nb = wx.aui.AuiNotebook(self, self._AUI_NOTEBOOK)
        welcomeMessage = wx.StaticText(self.nb, -1, _WelcomeMessage)
        self.nb.AddPage(welcomeMessage, "Welcome!")

        # Good thing the aui module breaks naming convention
        self.nb.Bind(wx.aui.EVT_AUINOTEBOOK_PAGE_CLOSE,
                     self.OnPageClosed)
        self.nb.Bind(wx.aui.EVT_AUINOTEBOOK_PAGE_CHANGED,
                     self.OnPageChanged)
        self.__set_properties()
        self.__do_layout()

    def OnPageClosed(self, event):
        deletedPage = event.Selection
        Publisher().sendMessage("page_deleted",deletedPage)

    def OnPageChanged(self, event):
        page = self.nb.GetPage(event.Selection)
        self.UpdateSubscriptions(page)
        Publisher().sendMessage("page_changed",page)

    def UpdateSubscriptions(self, page):
        for m,f in subscription_pairs:
            Publisher().unsubAll(("viewchange",m))
        if not hasattr(page,'CreationID'): return
        cti = page.CreationID
        toSubscribe = []
        if self._link_plots:
            for i in range(self.nb.GetPageCount()):
                p = self.nb.GetPage(i)
                try:
                    if p.CreationID == cti: toSubscribe.append(p)
                except AttributeError:
                    continue
        else: toSubscribe.append(page)
        for p in toSubscribe:
            Publisher().subscribe(self.MessageLogger)
            for m, f in subscription_pairs:
                ff = getattr(p,f)
                if ff is None: continue
                Publisher().subscribe(ff, ('viewchange',m))
        page.UpdateCanvas()

    def MessageLogger(self, message):
        print message

    def __set_properties(self):
        pass

    def __do_layout(self):
        PlotPanelSizer = wx.BoxSizer(wx.VERTICAL)
        PlotPanelSizer.Add(self.nb, 1, wx.EXPAND)
        self.SetSizer(PlotPanelSizer)
        self.Layout()

    def viewPF(self, event=None):
        r = ReasonParameterFileViewer(self, -1, outputfile=self.outputfile)
        r.Show()

    def OnSize(self, event):
        if self.CurrentlyResizing == True:
            return
        self.CurrentlyResizing = True
        #w,h = self.GetEffectiveMinSize()
        w,h = event.GetSize()
        xx = min(w,h)
        self.SetSize((xx,xx))
        w,h = self.vmPanel.GetClientSize()
        xx = min(w,h)
        xy = xx / self.vmPanel.figure.get_dpi()
        self.vmPanel.figure.set_size_inches(xy,xy)
        self.vmPanel.figure_canvas.SetSize((xx,xx))
        self.vmPanel.plot.redraw_image()
        self.Layout()
        self.CurrentlyResizing = False

    # For the following, we need to figure out which windows to apply the
    # action to before we apply it.
    # This is probably best accomplished through pubsub.

    def AddPlot(self, page, title, ID):
        self.nb.AddPage(page, title)
        wx.YieldIfNeeded()
        self.nb.SetSelection(self.nb.GetPageCount()-1)
        wx.YieldIfNeeded()
        # Now we subscribe to events with this ID

    def GetCurrentOutput(self):
        return self.nb.GetPage(self.nb.Selection).outputfile

    def OnCallSwitchField(self, event):
        self.nb.GetPage(self.nb.Selection).switch_field(event)

    def OnCallSetWidth(self, event):
        self.nb.GetPage(self.nb.Selection).set_width(event)

    def OnCallRedraw(self, event):
        self.nb.GetPage(self.nb.Selection).UpdateCanvas()

    def OnCallZoomTop(self, event):
        self.nb.GetPage(self.nb.Selection).fulldomain(event)

    def OnCallSetZLim(self, event):
        self.nb.GetPage(self.nb.Selection).set_zlim(event)

    def OnCallViewPF(self, event):
        of = self.GetCurrentOutput()

class PlotPage(wx.Panel):
    plot = None
    def __init__(self, parent, status_bar, mw=None, CreationID = -1,
                 parent_id = None):
        wx.Panel.__init__(self, parent)
        self.Hide()

        self.CreationID = CreationID
        self.parent = parent
        self.mw = mw
        self.parent_id = parent_id

        self.figure = be.matplotlib.figure.Figure()
        self.axes = self.figure.add_subplot(111, aspect='equal')
        self.status_bar = status_bar

        self.SetupControls()
        self.SetupFigure()
        self.SetupMenu()
        self.DoLayout()
        self.Show()

    def set_zlim(self, *args):
        zmin, zmax = Toolbars.ChooseLimits(self.plot)
        #self.ChangeLimits(zmin, zmax)
        Publisher().sendMessage(("viewchange","limits"),(zmin,zmax))
        self.UpdateCanvas(only_fig=True)

    def redraw(self, *args):
        self.UpdateCanvas()

    # These should be done with better DRY style
    def OnEditTitle(self, event):
        curTitle = self.figure.axes[-1].title.get_text()
        dlg = wx.TextEntryDialog(
                self, 'New title?',
                'Change Title?', curTitle)
        if dlg.ShowModal() == wx.ID_OK:
            self.figure.axes[0].set_title(dlg.GetValue())
        dlg.Destroy()
        self.figure_canvas.draw()

    def OnEditLegend(self, event):
        try:
            curLegend = self.plot.colorbar.ax.yaxis.get_label().get_text()
        except:
            curLegend = ""
        dlg = wx.TextEntryDialog(
                self, 'New legend?',
                'Change Legend?', curLegend)
        if dlg.ShowModal() == wx.ID_OK:
            self.plot.colorbar.set_label(dlg.GetValue())
        dlg.Destroy()
        self.figure_canvas.draw()

    def OnEditXLabel(self, event):
        try:
            curXLabel = self.plot.axis.xaxis.get_label().get_text()
        except:
            curXLabel = ""
        dlg = wx.TextEntryDialog(
                self, 'New x-label?',
                'Change x-label?', curXLabel)
        if dlg.ShowModal() == wx.ID_OK:
            self.plot.axes.set_xlabel(dlg.GetValue())
        dlg.Destroy()
        self.figure_canvas.draw()

    def OnEditYLabel(self, event):
        try:
            curYLabel = self.plot.axis.yaxis.get_label().get_text()
        except:
            curYLabel = ""
        dlg = wx.TextEntryDialog(
                self, 'New y-label?',
                'Change y-label?', curYLabel)
        if dlg.ShowModal() == wx.ID_OK:
            self.plot.axes.set_ylabel(dlg.GetValue())
        dlg.Destroy()
        self.figure_canvas.draw()

    def OnShowContextMenu(self, event):
        pos = event.GetPosition()
        pos = self.figure_canvas.ScreenToClient(pos)
        self._context_menu_position = pos
        self.figure_canvas.PopupMenu(self.popupmenu)

    def OnColorMapChoice(self, event):
        item = self.cmapmenu.FindItemById(event.GetId())
        text = item.GetText()
        Publisher().sendMessage(("viewchange","cmap"),text)

    def ChangeLimits(self, zmin, zmax):
        self.plot.set_zlim(zmin,zmax)
        self.figure_canvas.draw()
        # We don't call update canvas

    def ChangeColorMapFromMessage(self, message):
        cmap = message.data
        self.plot.set_cmap(cmap)
        self.UpdateCanvas()

    def ChangeLimitsFromMessage(self, message):
        zmin, zmax = message.data
        self.ChangeLimits(zmin,zmax)
        #self.UpdateCanvas()

    def ChangeFieldFromMessage(self, message):
        pass

    def ChangeCenterFromMessage(self, message):
        pass

    def ChangeFieldParamFromMessage(self, message):
        pass

    def WipePlotDataFromMessage(self, message):
        pass

    def ChangeWidthFromMessage(self, message):
        pass

    def SetupMenu(self):
        self.popupmenu = wx.Menu()
        self.cmapmenu = wx.Menu()
        self.editprops = wx.Menu()
        self.popupmenu.AppendMenu(-1, "Color Map", self.cmapmenu)
        cmapnames = sorted(color_maps)
        for cmap in cmapnames:
            item = self.cmapmenu.AppendRadioItem(-1, cmap)
            self.Bind(wx.EVT_MENU, self.OnColorMapChoice, item)
        self.popupmenu.AppendMenu(-1, "Edit Properties", self.editprops)
        dd = [('Edit Legend',self.OnEditLegend),
              ('Edit Title',self.OnEditTitle),
              ('Edit x-label',self.OnEditXLabel),
              ('Edit y-label',self.OnEditYLabel)]
        for title, func in dd:
            ii = self.editprops.Append(-1, title)
            self.Bind(wx.EVT_MENU, func, ii)
        # Now we pre-select the current cmap
        try:
            ccmap_name = self.plot.colorbar.cmap.name
        except AttributeError:
            ccmap_name = "jet"
        try:
            self.cmapmenu.FindItemById(self.cmapmenu.FindItem(ccmap_name)).Check(True)
        except AttributeError:
            pass

    def makePlot(self):
        pass

    def UpdateStatusBar(self, event):
        pass

    def ClickProcess(self, event):
        pass

    def SaveImage(self):
        self.plot._generate_prefix('Plot')
        wildcard = "PNG (*png)|*.png"
        dlg = wx.FileDialog( \
            self, message="Save Image As ...", defaultDir=os.getcwd(), \
            defaultFile=self.plot.prefix, wildcard=wildcard, style=wx.SAVE)
        if dlg.ShowModal() == wx.ID_OK:
            path = str(dlg.GetPath())
            orig_size = self.figure.get_size_inches()
            self.figure.set_size_inches((10,8))
            width = self.figure.get_dpi()*8.0
            buff = self.plot._get_buff(width=width)
            self.plot._axes.images[0].set_data(buff)
            self.plot._axes.set_xlim(-0.5,width-0.5)
            self.plot._axes.set_ylim(-0.5,width-0.5)
            self.UpdateCanvas(only_fig=True)
            self.figure_canvas.print_figure(path,format='png',
                                            dpi=self.figure.get_dpi())
            # We leave the higher-resolution buffer in there
            self.figure.savefig(path,format='png')
            self.figure.set_size_inches(orig_size)
            self.UpdateCanvas(only_fig=True)
        dlg.Destroy()

    def SetupFigure(self):
        self.makePlot()
        be.Initialize(canvas=FigureCanvas)
        self.figure_canvas = be.engineVals["canvas"](self, -1, self.figure)
        self.figure_canvas.mpl_connect('motion_notify_event', self.UpdateStatusBar)
        self.figure_canvas.Bind(wx.EVT_LEFT_DOWN, self.ClickProcess)
        self.figure_canvas.Bind(wx.EVT_CONTEXT_MENU, self.OnShowContextMenu)
        self.figure_canvas.Bind(wx.EVT_LEFT_DCLICK, self.OnShowContextMenu)

class VMPlotPage(PlotPage):
    def __init__(self, parent, status_bar, outputfile, axis, field="Density",
                 weight_field = None, mw=None, CreationID = -1,
                 center = None, parent_id = None):
        self.outputfile = outputfile
        self.field = field
        self.weight_field = weight_field
        self.axis = axis
        if center is None:
            center = [0.501, 0.501, 0.501]
        self.center = center
        self.AmDrawingCircle = False
        self.circles = []

        if ytcfg.getboolean("reason","centeronmax"):
            self.center = outputfile.hierarchy.find_max("Density")[1]

        PlotPage.__init__(self, parent, status_bar, mw, CreationID, parent_id)
        self.SetBackgroundColour(wx.NamedColor("WHITE"))
        self.UpdateWidth()

    def SetupMenu(self):
        PlotPage.SetupMenu(self)
        self.popupmenu.AppendSeparator()
        self.center_on_max = self.popupmenu.Append(-1, "Center on max")
        self.center_here = self.popupmenu.Append(-1, "Center here")
        self.popupmenu.AppendSeparator()
        self.grid_boundaries = self.popupmenu.AppendCheckItem(-1, "Show Grid Boundaries")
        self.grid_boundaries.Check(False)
        self.particles = self.popupmenu.AppendCheckItem(-1, "Show particles")
        self.particles.Check(False)
        self.velocities = self.popupmenu.AppendCheckItem(-1, "Show Velocities")
        self.velocities.Check(False)
        self.popupmenu.AppendSeparator()
        self.take_log_menu = self.popupmenu.AppendCheckItem(-1, "Take Log")
        self.take_log_menu.Check(self.plot.log_field)
        fullDomain = self.popupmenu.Append(-1, "Zoom Top")
        bulk_velocity = self.popupmenu.Append(-1, "Set Bulk Velocity")
        centered_sphere = self.popupmenu.Append(-1, "Get Sphere")

        self.Bind(wx.EVT_MENU, self.OnCenterOnMax, self.center_on_max)
        self.Bind(wx.EVT_MENU, self.OnCenterHere, self.center_here)
        self.Bind(wx.EVT_MENU, self.show_grid_boundaries, self.grid_boundaries)
        self.Bind(wx.EVT_MENU, self.show_particles, self.particles)
        self.Bind(wx.EVT_MENU, self.show_velocities, self.velocities)
        self.Bind(wx.EVT_MENU, self.take_log, self.take_log_menu)
        self.Bind(wx.EVT_MENU, self.fulldomain, fullDomain)
        self.Bind(wx.EVT_MENU, self.set_bulk_velocity, bulk_velocity)
        self.Bind(wx.EVT_MENU, self.get_centered_sphere, centered_sphere)

    def SetupFigure(self):
        PlotPage.SetupFigure(self)

        self.axes.set_xticks(())
        self.axes.set_yticks(())
        self.axes.set_ylabel("")
        self.axes.set_xlabel("")

    def SetupControls(self):

        self.widthSlider = wx.Slider(self, -1, wx.SL_HORIZONTAL | wx.SL_AUTOTICKS)
        self.vals = na.logspace(log10(25*self.outputfile.hierarchy.get_smallest_dx()),0,201)
        self.widthSlider.SetRange(0, 200)
        self.widthSlider.SetTickFreq(1,1)
        self.widthSlider.SetValue(200)

        self.widthBox = wx.TextCtrl(self, style=wx.TE_PROCESS_ENTER)
        self.widthBox.SetValue("1.0")

        self.choices = self.outputfile.units.keys()
        self.choices.sort()

        self.unitList = wx.Choice(self, choices=self.choices)

        self.unitList.Bind(wx.EVT_CHOICE, self.UpdateUnit)
        self.widthSlider.Bind(wx.EVT_SCROLL, self.UpdateTextFromScrollEvent)
        self.widthSlider.Bind(wx.EVT_SCROLL_THUMBRELEASE, self.UpdateWidth)
        self.widthBox.Bind(wx.EVT_TEXT_ENTER, self.UpdateWidthFromText)

        self.unitList.SetSelection(0)

    def DoLayout(self):

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.figure_canvas, 1, wx.EXPAND)

        self.ControlSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer.Add(self.ControlSizer, 0, wx.EXPAND)

        self.InputControl = wx.BoxSizer(wx.VERTICAL)
        self.InputControl.AddSpacer(10)
        self.InputControl.Add(self.widthBox, 0, 0, 0)
        self.InputControl.AddSpacer(10)
        self.InputControl.Add(self.unitList, 0, wx.EXPAND, 0)
        self.InputControl.AddSpacer(10)

        self.ControlSizer.AddSpacer(10)
        self.ControlSizer.Add(self.InputControl, 0, 0, 0)
        self.ControlSizer.AddSpacer(10)
        self.SliderSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.SliderSizer.Add(self.widthSlider, 1,  wx.ALIGN_CENTER, 0)
        self.ControlSizer.Add(self.SliderSizer, 1, wx.GROW | wx.ALIGN_CENTER, 0)
        self.ControlSizer.AddSpacer(10)

        self.SetSizer(self.sizer)
        self.Fit()

    def ChangeColorMapFromMessage(self, message):
        PlotPage.ChangeColorMapFromMessage(self, message)
        self.UpdateWidth()

    def OnCenterOnMax(self, event):
        v, c = self.outputfile.h.find_max("Density")
        Publisher().sendMessage(('viewchange','center'), c)
        self.UpdateWidth()

    _grid_boundaries_cbid = None
    def show_grid_boundaries(self, event):
        if self._grid_boundaries_cbid is not None:
            self.plot.remove_callback(self._grid_boundaries_cbid)
            self._grid_boundaries_cbid = None
        else:
            self._grid_boundaries_cbid = \
                self.plot.add_callback(raven.GridBoundaryCallback())
        self.UpdateCanvas()

    _velocities_cbid = None
    def show_velocities(self, event):
        if self._velocities_cbid is not None:
            self.plot.remove_callback(self._velocities_cbid)
            self._velocities_cbid = None
        else:
            xv = "%s-velocity" % (lagos.axis_names[lagos.x_dict[self.axis]])
            yv = "%s-velocity" % (lagos.axis_names[lagos.y_dict[self.axis]])
            self._velocities_cbid = \
                self.plot.add_callback(raven.QuiverCallback(xv,yv,20))
        self.UpdateCanvas()

    _particles_cbid = None
    def show_particles(self, event):
        if self._particles_cbid is not None:
            self.plot.remove_callback(self._particles_cbid)
            self._particles_cbid = None
        else:
            self._particles_cbid = \
                self.plot.add_callback(raven.ParticleCallback(self.axis, 1.0))
        self.UpdateCanvas()

    def OnCenterHere(self, event):
        xp, yp = self._context_menu_position
        x, y = self.ConvertPositionToDataPosition(xp, yp)
        if x == None or y == None: return
        self.ChangeCenter(x,y)
        self.UpdateWidth()

    def ConvertPositionToDataPosition(self, xp, yp):
        #if not self.figure.axes[0].in_axes(xp,yp): return None, None
        #xp, yp = self.figure.axes[0].transData.inverse_xy_tup((xp,yp))
        dx = (self.plot.xlim[1] - self.plot.xlim[0])/self.plot.pix[0]
        dy = (self.plot.ylim[1] - self.plot.ylim[0])/self.plot.pix[1]
        l, b, width, height = raven.PlotTypes._get_bounds(self.figure.axes[0].bbox)
        x = self.plot.xlim[0] + (dx * (xp-l))
        y = self.plot.ylim[1] - (dy * (yp-b))
        return x, y

    def ChangeCenter(self, x, y):
        newCenter = self.center[:]
        newCenter[lagos.x_dict[self.axis]] = x
        newCenter[lagos.y_dict[self.axis]] = y
        Publisher().sendMessage(('viewchange','center'), tuple(newCenter))
        self.UpdateWidth()

    def ClickProcess(self, event):
        xp, yp = event.X, event.Y
        if event.AltDown():
            x,y = self.ConvertPositionToDataPosition(xp, yp)
            self.ChangeCenter(x, y)
        elif event.ShiftDown():
            if self.AmDrawingCircle:
                dx = (self.plot.xlim[1] - self.plot.xlim[0])/self.plot.pix[0]
                dy = (self.plot.ylim[1] - self.plot.ylim[0])/self.plot.pix[1]
                dc=wx.ClientDC(self.figure_canvas)
                b=dc.GetBrush()
                b.SetStyle(wx.TRANSPARENT)
                dc.SetBrush(b)
                rp = sqrt((xp-self.x1)**2.0 + \
                          (yp-self.y1)**2.0)
                r = sqrt(((xp-self.x1)*dx)**2.0 + \
                         ((yp-self.y1)*dy)**2.0)
                dc.DrawCircle(self.x1,self.y1,rp)
                unitname = self.choices[self.unitList.GetSelection()]
                unit = self.outputfile[unitname]
                cc = self.center[:]
                xd,yd = self.ConvertPositionToDataPosition(self.x1, self.y1)
                cc[lagos.x_dict[self.axis]] = xd
                cc[lagos.y_dict[self.axis]] = yd
                sphere = self.outputfile.hierarchy.sphere( \
                    cc, r)
                self.mw._add_sphere("Sphere: %0.3e %s" \
                        % (r*unit, unitname), sphere,
                        self.parent_id)
                self.AmDrawingCircle = False
            else:
                self.x1 = xp
                self.y1 = yp
                self.AmDrawingCircle = True

    def ChangeWidthFromMessage(self, message):
        w,u=message.data
        # Now we need to update the text, units and bar
        wI = self.unitList.GetItems().index(u)
        self.unitList.SetSelection(wI)
        self.widthBox.SetValue("%0.5e" % (w))
        val = w/self.outputfile[u]
        dx = (log10(self.vals[0])-log10(self.vals[-1]))/201
        self.widthSlider.SetValue(200-int(log10(val)/dx))
        self.ChangeWidth(w,u)

    def ChangeWidth(self, width, unit):
        self.plot.set_width(width, unit)
        self.UpdateCanvas(only_fig=True)

    def UpdateUnit(self, event=None):
        pos = self.widthSlider.GetValue()
        self.UpdateTextFromScroll(pos)

    def UpdateTextFromScrollEvent(self, event):
        self.UpdateTextFromScroll(event.GetPosition())

    def UpdateTextFromScroll(self, pos):
        unitname = self.choices[self.unitList.GetSelection()]
        unit = self.outputfile[unitname]
        val = self.vals[pos] * unit
        self.widthBox.SetValue("%0.5e" % (val))

    def UpdateWidthFromText(self, event):
        val = float(event.GetString())
        if val < 0:
            return
        unitname = self.choices[self.unitList.GetSelection()]
        unit = self.outputfile[unitname]
        # Now we figure out the closest slider tick
        dx = (log10(self.vals[0])-log10(self.vals[-1]))/201
        self.widthSlider.SetValue(200-int(log10(val)/dx))
        Publisher().sendMessage(('viewchange','width'), (val,unitname))

    def UpdateWidth(self, pos = None):
        val = float(self.widthBox.GetValue())
        unitname = self.choices[self.unitList.GetSelection()]
        unit = self.outputfile[unitname]
        Publisher().sendMessage(('viewchange','width'), (val,unitname))

    def ChangeCursor(self, event):
        self.figure_canvas.SetCursor(wx.StockCursor(wx.CURSOR_BULLSEYE))

    def UpdateStatusBar(self, event):
        if event.inaxes:
            xp, yp = event.xdata, event.ydata
            dx = abs(self.plot.xlim[0] - self.plot.xlim[1])/self.plot.pix[0]
            dy = abs(self.plot.ylim[0] - self.plot.ylim[1])/self.plot.pix[1]
            x = (dx * xp) + self.plot.xlim[0]
            y = (dy * yp) + self.plot.ylim[0]
            xn = lagos.axis_names[lagos.x_dict[self.axis]]
            yn = lagos.axis_names[lagos.y_dict[self.axis]]
            self.status_bar.SetStatusText("%s = %0.12e" % (xn,x), 0)
            self.status_bar.SetStatusText("%s = %0.12e" % (yn,y), 1)
            self.status_bar.SetStatusText("v = %0.12e" % \
                                        (self.GetDataValue(xp,yp)), 2)

    def GetDataValue(self, x, y):
        return self.plot.image._A[int(y), int(x)]

    def take_log(self, event):
        self.plot.set_log_field(self.take_log_menu.IsChecked())
        self.UpdateCanvas()

    def fulldomain(self, *args):
        self.ChangeWidth(1,'1')
        Publisher().sendMessage(('viewchange','width'), (1,'1'))

    def set_bulk_velocity(self, *args):
        bv = Toolbars.GetBulkVelocity(self.outputfile, self.center)
        if bv is not None:
            Publisher().sendMessage(('viewchange','field_param'),
                                    ('bulk_velocity', bv))
            Publisher().sendMessage(('viewchange','wipe'),())

    def get_centered_sphere(self, *args):
        rv = Toolbars.GetSphereRadius(self.outputfile, self.center)
        if rv is not None:
            radius, unit = rv
            sphere = self.outputfile.hierarchy.sphere( \
                self.center,
                radius/self.outputfile[unit])
            self.mw._add_sphere("Sphere: %0.3e %s" \
                    % (radius, unit), sphere, self.parent_id)

    def WipePlotDataFromMessage(self, message):
        self.data.clear_data()
        self.UpdateCanvas()

    def ChangeFieldParamFromMessage(self, message):
        param, val = message.data
        self.data.set_field_parameter(param, val)

    def set_width(self, *args):
        w, u = Toolbars.ChooseWidth(self.outputfile)
        #self.ChangeWidth(w,u) # Do we need this?
        Publisher().sendMessage(('viewchange','width'), (w,u))

    def ChangeFieldFromMessage(self, message):
        field = message.data
        self.ChangeField(field)

    def ChangeField(self, field):
        self.plot.switch_z(field)
        self.UpdateCanvas()
        self.take_log_menu.Check(self.plot.log_field)

    def switch_field(self, event):
        field = event.GetString()
        self.ChangeField(field)
        Publisher().sendMessage(('viewchange','field'), field)

    def UpdateCanvas(self, only_fig=False):
        if self.IsShown():
            if not only_fig: self.plot._redraw_image()
            self.figure_canvas.draw()

    def ChangeCenterFromMessage(self, message):
        x, y, z = message.data
        # We are dealing with a pass-by-reference center
        self.center[0] = x
        self.center[1] = y
        self.center[2] = z
        self.data.set_field_parameter('center',[x,y,z])
        #self.UpdateWidth()

    def QueryFields(self):
        return QueryFields(self.outputfile, True)

class SlicePlotPage(VMPlotPage):
    def makePlot(self):
        self.data = self.outputfile.hierarchy.slice(self.axis,
                                    self.center[self.axis], self.field, self.center)
        self.plot = be.SlicePlot(self.data, self.field, figure=self.figure, axes=self.axes)

    def ChangeCenterFromMessage(self, message):
        VMPlotPage.ChangeCenterFromMessage(self, message)
        self.data.reslice(self.center[self.axis])


class ProjPlotPage(VMPlotPage):
    def makePlot(self):
        self.data = self.outputfile.hierarchy.proj(self.axis,
                                   self.field, weight_field=self.weight_field,
                                   center=self.center)
        self.plot = be.ProjectionPlot(self.data, self.field, figure=self.figure,
                                      axes=self.axes)

    def QueryFields(self):
        return [self.field]

class CuttingPlanePlotPage(VMPlotPage):
    def __init__(self, *args, **kwargs):
        self.normal = kwargs.pop("normal")
        VMPlotPage.__init__(self, *args, **kwargs)
        self.__fix_menu()

    def makePlot(self):
        self.data = self.outputfile.hierarchy.cutting(self.normal,
                                    self.center, self.field)
        self.plot = be.CuttingPlanePlot(self.data, self.field,
                                    figure=self.figure, axes=self.axes)

    def UpdateStatusBar(self, event):
        pass

    def __fix_menu(self):
        self.center_on_max.Enable(False)
        self.center_here.Enable(False)
        self.grid_boundaries.Enable(False)
        self.particles.Enable(False)

    def show_velocities(self, event):
        if self._velocities_cbid is not None:
            self.plot.remove_callback(self._velocities_cbid)
            self._velocities_cbid = None
        else:
            xv = "CuttingPlaneVelocityX"
            yv = "CuttingPlaneVelocityY"
            self._velocities_cbid = \
                self.plot.add_callback(raven.CuttingQuiverCallback(xv,yv,16))
        self.UpdateCanvas()

class ProfilePlotPage(PlotPage):
    def __init__(self, parent, status_bar, data_object, argdict, mw=None,
                 CreationID = -1, parent_id = None):
        self.data_object = data_object
        self.argdict = argdict

        PlotPage.__init__(self, parent, status_bar, mw, CreationID, parent_id)
        self.UpdateAvailableFields()

    def SetupControls(self):
        self.ButtonPanel = wx.Panel(self, -1)
        self.sbox = wx.StaticBox(self.ButtonPanel, -1, "Field Selection")
        self.z_field = wx.Choice(self.ButtonPanel, -1, choices=[])
        self.add_field = wx.Button(self.ButtonPanel, -1, "Add Field")
        self.Bind(wx.EVT_BUTTON, self.AddField, self.add_field)
        self.Bind(wx.EVT_CHOICE, self.switch_z, self.z_field)

    def SetupMenu(self):
        PlotPage.SetupMenu(self)
        self.popupmenu.AppendSeparator()
        self.take_log_menu = self.popupmenu.AppendCheckItem(-1, "Take Log")

        self.Bind(wx.EVT_MENU, self.take_log, self.take_log_menu)

    def DoLayout(self):
        self.MainSizer = wx.BoxSizer(wx.VERTICAL)
        self.MainSizer.Add(self.figure_canvas, 1, wx.EXPAND)
        self.MainSizer.Add(self.ButtonPanel, 0, wx.EXPAND)

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(self.z_field, 1, wx.EXPAND)
        sizer.Add(self.add_field, 0, wx.EXPAND)
        self.ButtonPanel.SetSizer(sizer)
        self.ButtonPanel.Layout()

        self.SetSizer(self.MainSizer)
        self.Layout()

    def UpdateAvailableFields(self):
        self.z_field.SetItems(sorted(self.QueryAvailableFields()))

    def switch_z(self, event=None):
        if self.plot is None: return
        self.plot.switch_z(self.z_field.GetStringSelection())
        self.UpdateCanvas()

    def UpdateCanvas(self, only_fig=False):
        if self.IsShown():
            if not only_fig: self.plot._redraw_image()
            self.figure_canvas.draw()

    def take_log(self, event):
        self.plot.set_log_field(self.take_log_menu.IsChecked())
        self.UpdateCanvas()

class Profile1DPlotPage(ProfilePlotPage):
    def AddField(self, event=None):
        pf1daf = Profile1DAddField(self.data_object, self)
        if pf1daf.ShowModal() == wx.ID_OK:
            argdict = pf1daf.return_argdict()
            if argdict is None:
                pf1daf.Destroy()
                self.UpdateAvailableFields()
                return
            self.data.add_fields(**argdict)
            self.UpdateAvailableFields()
            self.z_field.SetSelection(
                    self.z_field.GetItems().index(argdict['fields']))
            self.switch_z()
        pf1daf.Destroy()

    def SetupMenu(self):
        ProfilePlotPage.SetupMenu(self)
        self.take_log_menu.Check(self.plot._log_y)

    def QueryAvailableFields(self):
        fields = []
        for field in self.data.keys():
            if field != self.data.bin_field:
                fields.append(field)
        return fields

    def makePlot(self, event=None):
        self.data = lagos.BinnedProfile1D(self.data_object, **self.argdict)
        self.data.pf = self.data_object.pf
        self.data.hierarchy = self.data_object.pf.h
        # Now, unfortunately, we prompt for the field
        self.AddField(None)
        self.plot = be.Profile1DPlot(self.data, 
                                [self.argdict['bin_field'],
                                 self.z_field.GetStringSelection()],
                                 id = wx.NewId(),
                                 figure=self.figure, axes=self.axes)

    def UpdateStatusBar(self, event):
        pass

class PhasePlotPage(ProfilePlotPage):
    def AddField(self, event=None):
        pf2daf = Profile2DAddField(self.data_object, self)
        if pf2daf.ShowModal() == wx.ID_OK:
            argdict = pf2daf.return_argdict()
            if argdict is None:
                pf2daf.Destroy()
                self.UpdateAvailableFields()
                return
            self.data.add_fields(**argdict)
            self.UpdateAvailableFields()
            self.z_field.SetSelection(
                    self.z_field.GetItems().index(argdict['fields']))
            self.switch_z()
        pf2daf.Destroy()

    def SetupMenu(self):
        ProfilePlotPage.SetupMenu(self)
        self.take_log_menu.Check(self.plot._log_z)

    def makePlot(self, event=None):
        self.data = lagos.BinnedProfile2D(self.data_object, **self.argdict)
        self.data.pf = self.data_object.pf
        self.data.hierarchy = self.data_object.pf.h
        # Now, unfortunately, we prompt for the field
        self.AddField(None)
        self.plot = be.PhasePlot(self.data, 
                                [self.argdict['x_bin_field'],
                                 self.argdict['y_bin_field'],
                                 self.z_field.GetStringSelection()],
                                 id = wx.NewId(),
                                 figure=self.figure, axes=self.axes)

    def UpdateStatusBar(self, event):
        if event.inaxes:
            if not hasattr(self.plot, 'pix'): return
            xp, yp = event.xdata, event.ydata
            xn = self.data.x_bin_field
            yn = self.data.y_bin_field
            vn = self.FieldZ.StringSelection
            self.status_bar.SetStatusText("%s = %0.5e" % (xn, xp), 0)
            self.status_bar.SetStatusText("%s = %0.5e" % (yn, yp), 1)
            self.status_bar.SetStatusText("%s = %0.5e" % \
                                        (vn, self.GetDataValue(xp,yp)), 2)

    def GetDataValue(self, x, y):
        xi = na.digitize(na.array([x]), self.plot.x_bins)-1
        yi = na.digitize(na.array([y]), self.plot.y_bins)-1
        return self.plot.vals[yi, xi]

    def QueryAvailableFields(self):
        fields = []
        for field in self.data.keys():
            if field != self.data.x_bin_field and \
               field != self.data.y_bin_field:
                fields.append(field)
        return fields

    def ChangeLimits(self, zmin, zmax):
        self.plot.set_zlim(zmin,zmax)
        self.figure_canvas.draw()
        # We don't call update canvas

if __name__ == "__main__":
    app = ReasonApp(0)
    app.MainLoop()



"""
A collection of helper functions, most generally for things
that SciPy doesn't have that I expected it to

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2007-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.lagos import *

class UnilinearFieldInterpolator:
    def __init__(self, table, boundaries, field_names, truncate=False):
        self.table = table.astype('float64')
        self.truncate = truncate
        x0, x1 = boundaries
        self.x_name = field_names
        self.x_bins = na.linspace(x0, x1, table.shape[0]).astype('float64')

    def __call__(self, data_object):
        orig_shape = data_object[self.x_name].shape
        x_vals = data_object[self.x_name].ravel().astype('float64')

        x_i = (na.digitize(x_vals, self.x_bins) - 1).astype('int32')
        if na.any((x_i == -1) | (x_i == len(self.x_bins)-1)):
            if not self.truncate:
                mylog.error("Sorry, but your values are outside" + \
                            " the table!  Dunno what to do, so dying.")
                mylog.error("Error was in: %s", data_object)
                raise ValueError
            else:
                x_i = na.minimum(na.maximum(x_i,0), len(self.x_bins)-2)

        my_vals = na.zeros(x_vals.shape, dtype='float64')
        amr_utils.UnilinearlyInterpolate(self.table, x_vals, self.x_bins, x_i, my_vals)
        return my_vals.reshape(orig_shape)

class BilinearFieldInterpolator:
    def __init__(self, table, boundaries, field_names, truncate=False):
        self.table = table.astype('float64')
        self.truncate = truncate
        x0, x1, y0, y1 = boundaries
        self.x_name, self.y_name = field_names
        self.x_bins = na.linspace(x0, x1, table.shape[0]).astype('float64')
        self.y_bins = na.linspace(y0, y1, table.shape[1]).astype('float64')

    def __call__(self, data_object):
        orig_shape = data_object[self.x_name].shape
        x_vals = data_object[self.x_name].ravel().astype('float64')
        y_vals = data_object[self.y_name].ravel().astype('float64')

        x_i = (na.digitize(x_vals, self.x_bins) - 1).astype('int32')
        y_i = (na.digitize(y_vals, self.y_bins) - 1).astype('int32')
        if na.any((x_i == -1) | (x_i == len(self.x_bins)-1)) \
            or na.any((y_i == -1) | (y_i == len(self.y_bins)-1)):
            if not self.truncate:
                mylog.error("Sorry, but your values are outside" + \
                            " the table!  Dunno what to do, so dying.")
                mylog.error("Error was in: %s", data_object)
                raise ValueError
            else:
                x_i = na.minimum(na.maximum(x_i,0), len(self.x_bins)-2)
                y_i = na.minimum(na.maximum(y_i,0), len(self.y_bins)-2)

        my_vals = na.zeros(x_vals.shape, dtype='float64')
        amr_utils.BilinearlyInterpolate(self.table,
                                 x_vals, y_vals, self.x_bins, self.y_bins,
                                 x_i, y_i, my_vals)
        return my_vals.reshape(orig_shape)

class TrilinearFieldInterpolator:
    def __init__(self, table, boundaries, field_names, truncate = False):
        self.table = table.astype('float64')
        self.truncate = truncate
        x0, x1, y0, y1, z0, z1 = boundaries
        self.x_name, self.y_name, self.z_name = field_names
        self.x_bins = na.linspace(x0, x1, table.shape[0]).astype('float64')
        self.y_bins = na.linspace(y0, y1, table.shape[1]).astype('float64')
        self.z_bins = na.linspace(z0, z1, table.shape[2]).astype('float64')

    def __call__(self, data_object):
        orig_shape = data_object[self.x_name].shape
        x_vals = data_object[self.x_name].ravel().astype('float64')
        y_vals = data_object[self.y_name].ravel().astype('float64')
        z_vals = data_object[self.z_name].ravel().astype('float64')

        x_i = na.digitize(x_vals, self.x_bins) - 1
        y_i = na.digitize(y_vals, self.y_bins) - 1
        z_i = na.digitize(z_vals, self.z_bins) - 1
        if na.any((x_i == -1) | (x_i == len(self.x_bins)-1)) \
            or na.any((y_i == -1) | (y_i == len(self.y_bins)-1)) \
            or na.any((z_i == -1) | (z_i == len(self.z_bins)-1)):
            if not self.truncate:
                mylog.error("Sorry, but your values are outside" + \
                            " the table!  Dunno what to do, so dying.")
                mylog.error("Error was in: %s", data_object)
                raise ValueError
            else:
                x_i = na.minimum(na.maximum(x_i,0), len(self.x_bins)-2)
                y_i = na.minimum(na.maximum(y_i,0), len(self.y_bins)-2)
                z_i = na.minimum(na.maximum(z_i,0), len(self.z_bins)-2)

        my_vals = na.zeros(x_vals.shape, dtype='float64')
        amr_utils.TrilinearlyInterpolate(self.table,
                                 x_vals, y_vals, z_vals,
                                 self.x_bins, self.y_bins, self.z_bins,
                                 x_i, y_i, z_i, my_vals)
        return my_vals.reshape(orig_shape)

        # Use notation from Paul Bourke's page on interpolation
        # http://local.wasp.uwa.edu.au/~pbourke/other/interpolation/
        x = (x_vals - self.x_bins[x_i]) / (self.x_bins[x_i+1] - self.x_bins[x_i])
        y = (y_vals - self.y_bins[y_i]) / (self.y_bins[y_i+1] - self.y_bins[y_i])
        z = (z_vals - self.z_bins[z_i]) / (self.z_bins[z_i+1] - self.z_bins[z_i])
        xm = (self.x_bins[x_i+1] - x_vals) / (self.x_bins[x_i+1] - self.x_bins[x_i])
        ym = (self.y_bins[y_i+1] - y_vals) / (self.y_bins[y_i+1] - self.y_bins[y_i])
        zm = (self.z_bins[z_i+1] - z_vals) / (self.z_bins[z_i+1] - self.z_bins[z_i])
        if na.any(na.isnan(self.table)):
            raise ValueError
        if na.any(na.isnan(x) | na.isnan(y) | na.isnan(z)):
            raise ValueError
        if na.any(na.isnan(xm) | na.isnan(ym) | na.isnan(zm)):
            raise ValueError
        my_vals  = self.table[x_i  ,y_i  ,z_i  ] * (xm*ym*zm)
        my_vals += self.table[x_i+1,y_i  ,z_i  ] * (x *ym*zm)
        my_vals += self.table[x_i  ,y_i+1,z_i  ] * (xm*y *zm)
        my_vals += self.table[x_i  ,y_i  ,z_i+1] * (xm*ym*z )
        my_vals += self.table[x_i+1,y_i  ,z_i+1] * (x *ym*z )
        my_vals += self.table[x_i  ,y_i+1,z_i+1] * (xm*y *z )
        my_vals += self.table[x_i+1,y_i+1,z_i  ] * (x *y *zm)
        my_vals += self.table[x_i+1,y_i+1,z_i+1] * (x *y *z )
        return my_vals.reshape(orig_shape)

def get_centers(pf, filename, center_cols, radius_col, unit='1'):
    """
    Return an iterator over EnzoSphere objects generated from the appropriate 
    columns in *filename*.  Optionally specify the *unit* radius is in.
    """
    sp_list = []
    for line in open(filename):
        if line.startswith("#"): continue
        vals = line.split()
        x,y,z = [float(vals[i]) for i in center_cols]
        r = float(vals[radius_col])
        yield pf.h.sphere([x,y,z], r/pf[unit])



"""
A set of utilities for analyzing disks

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2008-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.mods import *

class StackedDiskImage(object):
    def __init__(self, pf, center, norm_vec, thickness, width,
                 nslices = 100, buff_size=(800,800)):
        """
        This class implements an AMR data object that will stack up
        oblique-slices to generate an image along an arbitrary axis.
        """
        self.data = {}
        self.pf = pf
        self.center = na.array(center)
        self.norm_vec = norm_vec/na.sqrt(na.dot(norm_vec,norm_vec))
        self.thickness = thickness
        self.width = width
        self.nslices = nslices
        self.buff_size = buff_size

    def __getitem__(self, key):
        if key not in self.data:
            self[key] = self.get_data(key)
        return self.data[key]

    def __delitem__(self, key):
        del self.data[key]

    def __setitem__(self, key, val):
        self.data[key] = val

    def __iter__(self):
        kk = self.data.keys()
        for key in kk: yield key

    def get_data(self, field):
        # Needs to be rewritten to get multiple fields
        buffer = na.zeros(self.buff_size, dtype='float64')
        wsp = (-self.width/2.0, self.width/2.0,
               -self.width/2.0, self.width/2.0)
        for d in na.mgrid[-self.thickness/2.0:self.thickness/2.0:self.nslices*1j]:
            print d, self.thickness/2.0
            center = self.center + self.norm_vec*d
            cut = self.pf.h.cutting(self.norm_vec, center)
            buff = raven.ObliqueFixedResolutionBuffer(
                        cut, wsp, self.buff_size)
            buffer += buff[field]
        return buffer
